"""
Taylor diagram (Taylor, 2001) test implementation.

http://www-pcmdi.llnl.gov/about/staff/Taylor/CV/Taylor_diagram_primer.htm
"""

__version__ = 'Time-stamp: <2012-02-17 20:59:35 ycopin>'
__author__ = 'Yannick Copin <yannick.copin@laposte.net>'

# Rev. 1.0  saeed Moghimi  moghimis@gmail.com

from matplotlib import pyplot
from matplotlib.projections import PolarAxes
from mpl_toolkits.axisartist import floating_axes, grid_finder
import numpy


class TaylorDiagram:
    """Taylor diagram: plot model standard deviation and correlation
    to reference (data) sample in a single-quadrant polar plot, with
    r=stddev and theta=arccos(correlation).
    """

    def __init__(self, refstd, fig=None, rect=111, label='_'):
        """Set up Taylor diagram axes, i.e. single quadrant polar
        plot, using mpl_toolkits.axisartist.floating_axes. refstd is
        the reference standard deviation to be compared to.
        """

        self.refstd = refstd  # Reference standard deviation

        tr = PolarAxes.PolarTransform()

        # Correlation labels
        rlocs = numpy.concatenate((numpy.arange(10) / 10.0, [0.95, 0.99]))
        tlocs = numpy.arccos(rlocs)  # Conversion to polar angles
        gl1 = grid_finder.FixedLocator(tlocs)  # Positions
        tf1 = grid_finder.DictFormatter(dict(zip(tlocs, map(str, rlocs))))

        # Standard deviation axis extent
        self.smin = 0
        self.smax = 1.5 * self.refstd

        ghelper = floating_axes.GridHelperCurveLinear(
            tr,
            extremes=(0, numpy.pi / 2, self.smin, self.smax),  # 1st quadrant
            grid_locator1=gl1,
            tick_formatter1=tf1,
        )

        if fig is None:
            fig = pyplot.figure()

        ax = floating_axes.FloatingSubplot(fig, rect, grid_helper=ghelper)
        fig.add_subplot(ax)

        # Adjust axes
        ax.axis['top'].set_axis_direction('bottom')  # "Angle axis"
        ax.axis['top'].toggle(ticklabels=True, label=True)
        ax.axis['top'].major_ticklabels.set_axis_direction('top')
        ax.axis['top'].label.set_axis_direction('top')
        ax.axis['top'].label.set_text('Correlation')

        ax.axis['left'].set_axis_direction('bottom')  # "X axis"
        ax.axis['left'].label.set_text('Standard deviation')

        ax.axis['right'].set_axis_direction('top')  # "Y axis"
        ax.axis['right'].label.set_text('Standard deviation')

        ax.axis['right'].toggle(ticklabels=True)
        ax.axis['right'].major_ticklabels.set_axis_direction('left')

        ax.axis['bottom'].set_visible(False)  # Useless

        # Contours along standard deviations
        ax.grid(False)

        self._ax = ax  # Graphical axes
        self.ax = ax.get_aux_axes(tr)  # Polar coordinates

        # Add reference point and stddev contour
        print('Reference std:', self.refstd)
        # l, = self.ax.plot([0], self.refstd, 'k*',     rem by saeed
        #                  ls='', ms=10, label=label)  rem by saeed
        # t = NP.linspace(0, NP.pi/2)                   rem by saeed
        # r = NP.zeros_like(t) + self.refstd            rem by saeed
        # self.ax.plot(t,r, 'k--', label='_')           rem by saeed

        for ip in range(len(rlocs)):
            x = [self.smin, self.smax]
            y = [numpy.arccos(rlocs[ip]), numpy.arccos(rlocs[ip])]
            self.ax.plot(y, x, 'grey', linewidth=0.25)

        # Collect sample points for latter use (e.g. legend)
        # self.samplePoints = [l]                       rem by saeed
        self.samplePoints = []  # add by saeed

    def add_sample(self, stddev, corrcoef, ref, *args, **kwargs):
        """Add sample (stddev,corrcoeff) to the Taylor diagram. args
        and kwargs are directly propagated to the Figure.plot
        command."""

        (l,) = self.ax.plot(numpy.arccos(corrcoef), stddev, *args, **kwargs)  # (theta,radius)
        self.samplePoints.append(l)

        if ref:
            t = numpy.linspace(0, numpy.pi / 2)  # add by saeed
            r = numpy.zeros_like(t) + stddev  # add by saeed
            self.ax.plot(t, r, 'grey', linewidth=0.25, label='_')  # add by saeed

        return l

    def add_contours(self, levels, data_std, **kwargs):
        """Add constant centered RMS difference contours."""

        rs, ts = numpy.meshgrid(
            numpy.linspace(self.smin, self.smax), numpy.linspace(0, numpy.pi / 2)
        )
        # Compute centered RMS difference
        rms = numpy.sqrt(data_std ** 2 + rs ** 2 - 2 * data_std * rs * numpy.cos(ts))

        contours = self.ax.contour(ts, rs, rms, levels, **kwargs)

        return contours



from abc import ABC, abstractmethod
from enum import Enum
from os import PathLike
from pathlib import Path
from typing import Any, Collection, Dict, List, Mapping, Tuple, Union
from functools import reduce
from copy import deepcopy
from collections import UserDict
import fnmatch

import dask
import geopandas
import f90nml
from geopandas import GeoDataFrame
import numpy
import pandas
from pandas import DataFrame
from pyproj.transformer import Transformer
from pyschism.mesh import Hgrid
from scipy.spatial import KDTree
from shapely.geometry import Point
from stormevents.nhc import VortexTrack
from typepigeon import convert_value
import xarray
from xarray import DataArray, Dataset

from ensembleperturbation.perturbation.atcf import parse_vortex_perturbations
from ensembleperturbation.utilities import get_logger


LOGGER = get_logger('parsing.schism')

SCHISM_ADCIRC_COORD_MAPPING = {
    'nSCHISM_hgrid_node': 'node',
    'SCHISM_hgrid_node_x': 'x',
    'SCHISM_hgrid_node_y': 'y',
}

SCHISM_ADCIRC_VAR_MAPPING = {
    'elevation': 'zeta',
    'horizontalVelX': 'u-vel',
    'horizontalVelY': 'v-vel',
    'windSpeedX': 'windx',
    'windSpeedY': 'windy',
    'airPressure': 'pressure',
    'max_elevation': 'zeta_max',
    'max_elevation_times': 'time_of_zeta_max',
    'max_velocity': 'vel_max',
    'max_velocity_times': 'time_of_vel_max',
    'min_pressure': 'pressure_min',
    'min_pressure_times': 'time_of_pressure_min',
    'max_wind': 'wind_max',
    'max_wind_times': 'time_of_wind_max',
    'station_index': 'station_name',
}

SCHISM_ADCIRC_OUT_MAPPING = {
    'schism_point_elevtion.nc': 'fort.61.nc',  # ['station_name', 'zeta']
    'schism_point_velocity.nc': 'fort.62.nc',  # ['station_name', 'u-vel', 'v-vel']
    'schism_max_elevation.nc': 'maxele.63.nc',  # ['zeta_max', 'time_of_zeta_max']
    'schism_max_velocity.nc': 'maxvel.63.nc',  # ['vel_max', 'time_of_vel_max']
    'schism_min_pressure.nc': 'minpr.63.nc',  # ['pressure_min', 'time_of_pressure_min']
    'schism_max_wind.nc': 'maxwvel.63.nc',  # ['wind_max', 'time_of_wind_max']
    'schism_elevation.nc': 'fort.63.nc',  # ['zeta']
    'schism_velocity.nc': 'fort.64.nc',  # ['u-vel', 'v-vel']
    'schism_pressure.nc': 'fort.73.nc',  # ['pressure']
    'schism_wind.nc': 'fort.74.nc',  # ['windx', 'windy']
}


def is_stacked(pattern):
    return '*' in pattern


def create_output_dict(file_pattern, directory=Path(), existing_dict=None):
    output_dict = {}
    if isinstance(existing_dict, Mapping):
        output_dict = deepcopy(existing_dict)

    matches = list(directory.glob(f'**/{file_pattern}'))
    proper_matches = [path for path in matches if path.parent.name == 'outputs']
    run_dirs = set(path.parent.parent for path in proper_matches)
    for rundir in run_dirs:
        run_outputs = [path for path in proper_matches if path.match(f'{rundir}/outputs/*')]
        n_stacks = 1
        if is_stacked(file_pattern):
            n_stacks = max(int(out.stem.split('_')[-1]) for out in run_outputs)
        run_dict = output_dict.setdefault(rundir, dict())
        run_dict['name'] = rundir.name
        run_out_dict = run_dict.setdefault('outputs', dict())
        pattern_dict = run_out_dict.setdefault(file_pattern, dict())
        pattern_dict['n_stacks'] = n_stacks
        pattern_dict['files'] = run_outputs

    return output_dict


def validate_run_output(run_dict, output_patterns):
    # Check if it has all patterns
    if not all(patt in run_dict['outputs'] for patt in output_patterns):
        return False

    # Check if num of stacks match, for stacked (in time) outputs
    if (
        len(
            {
                run_dict['outputs'][patt]['n_stacks']
                for patt in output_patterns
                if is_stacked(patt)
            }
        )
        > 1
    ):
        return False

    return True


def find_run_dir_for_output(file_patterns, directory=Path(), validate=True):

    output_dict = {}
    for pattern in file_patterns:
        output_dict = create_output_dict(pattern, directory, output_dict)

    if validate:
        validated_outputs = {}
        for key, run_dict in output_dict.items():
            if validate_run_output(run_dict, file_patterns):
                validated_outputs[key] = run_dict
        output_dict = validated_outputs

    if len(output_dict) > 0:
        LOGGER.info(
            f'found {len(output_dict)} run directories with all the specified output patterns'
        )
    else:
        raise FileNotFoundError(
            f'could not find any run directories with all the specified output patterns'
        )

    return output_dict


class ElevationSelection(Enum):
    wet = 'wet'
    inundated = 'inundated'
    dry = 'dry'


class SchismOutput(ABC):
    out_filename: str
    file_patterns: str
    variables: List[str]
    drop_variables: List[str] = []
    nodata: float = -99999.0

    @classmethod
    @abstractmethod
    def read(cls, filename: PathLike, names: List[str] = None) -> Union[DataFrame, DataArray]:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def read_directory(
        cls, directory: PathLike, variables: List[str] = None, parallel: bool = False
    ) -> Dataset:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def subset(
        cls,
        dataset: Union[Dataset, DataArray],
        bounds: (float, float, float, float) = None,
        **kwargs,
    ) -> Union[Dataset, DataArray]:
        raise NotImplementedError


class TimeSeriesOutput(ABC):
    pass


class StationTimeSeriesOutput(SchismOutput, TimeSeriesOutput, ABC):

    station_file = 'station.in'

    """
    outputs/staout_[1..,9], corresponding respectively to elev,
    air pressure, wind u, wind v, T, S, u, v, w
    """

    @classmethod
    def read(
        cls,
        filenames: Union[PathLike, List[PathLike]],
        station_file: PathLike,
        station_idx: List[int] = None,
    ) -> GeoDataFrame:

        if not isinstance(filenames, List):
            filenames = [filenames]

        if not all(any(j.match(i) for j in filenames) for i in cls.file_patterns):
            raise ValueError(
                f'{cls.__name__} requires files that match {" & ".join(cls.file_patterns)}'
                f' input files for {" & ".join(cls.variables)}'
            )

        station_df = (
            pandas.read_csv(
                station_file, header=None, delim_whitespace=True, index_col=0, skiprows=2
            )
            .drop(columns=[3, 4])
            .rename_axis('station_index')
            .rename(columns={1: 'x', 2: 'y'})
        )

        all_results_df = pandas.DataFrame()
        for fname, var in zip(filenames, cls.variables):
            result_df = (
                pandas.read_csv(fname, header=None, delim_whitespace=True, index_col=0)
                .transpose()
                .melt(var_name='time', ignore_index=False)
                .rename_axis('station_index')
                .rename(columns={'value': var})
            )
            result_df['x'] = station_df.loc[
                result_df.index.get_level_values('station_index')
            ].x
            result_df['y'] = station_df.loc[
                result_df.index.get_level_values('station_index')
            ].y
            result_df = result_df.set_index(['time', result_df.index])
            all_results_df = pandas.concat((all_results_df, result_df), axis=1)
        all_results_df = all_results_df.reset_index()

        # TODO: Use station_idx if none to only return stations of interest
        station_gdf = geopandas.GeoDataFrame(
            geometry=geopandas.points_from_xy(
                station_df.loc[all_results_df.station_index].x,
                station_df.loc[all_results_df.station_index].y,
            ),
            data=all_results_df,
        )

        return station_gdf

    @classmethod
    def read_to_dataset(
        cls,
        filenames: Union[PathLike, List[PathLike]],
        station_file: PathLike,
        station_idx: List[int] = None,
        run_name: str = None,
    ) -> Dataset:

        station_gdf = cls.read(filenames, station_file, station_idx)
        if run_name is not None:
            station_gdf['run'] = run_name

        dataset = xarray.Dataset.from_dataframe(
            station_gdf.set_index(['run', 'time', 'station_index'])
        )
        dataset = dataset.assign(
            {
                'geometry': dataset.geometry.isel(run=0, time=0),
                'x': dataset.x.isel(run=0, time=0),
                'y': dataset.y.isel(run=0, time=0),
            }
        )

        return dataset

    @classmethod
    def read_directory(
        cls, directory: PathLike, variables: List[str] = None, parallel: bool = False
    ) -> GeoDataFrame:
        """
        Compile a dataset from output files in the given directory.
        This could be called for a single run directory or a parent
        of multiple run directories.

        :param directory: directory containing output files
        :param parallel: load data concurrently with Dask
        :return: GeoDataFrame of output data
        """

        if not isinstance(directory, Path):
            directory = Path(directory)

        output_dict = find_run_dir_for_output(cls.file_patterns, directory)
        run_dirs = list(output_dict.keys())

        # TODO: Use variables?

        dataset_list = []
        lazy_results = []
        for en, run_dir in enumerate(run_dirs):
            if parallel:
                lazy_results.append(
                    dask.delayed(cls.read_to_dataset)(
                        [
                            f
                            for patt_dict in output_dict[run_dir]['outputs'].values()
                            for f in patt_dict['files']
                        ],
                        run_dir / cls.station_file,
                        run_name=run_dir.name,
                    )
                )

            else:
                dataset_list.append(
                    cls.read_to_dataset(
                        [
                            f
                            for patt_dict in output_dict[run_dir]['outputs'].values()
                            for f in patt_dict['files']
                        ],
                        run_dir / cls.station_file,
                        run_name=run_dir.name,
                    )
                )

        if len(lazy_results) > 0:
            dataset_list = dask.compute(*lazy_results)

        return xarray.merge(dataset_list)

    @classmethod
    def subset(
        cls,
        dataset: Union[Dataset, DataArray],
        bounds: (float, float, float, float) = None,
        **kwargs,
    ) -> Union[Dataset, DataArray]:

        subset = ~dataset['station_index'].isnull()

        if bounds is not None:
            LOGGER.debug(f'filtering within bounds {bounds}')
            if bounds[0] is not None:
                subset = numpy.logical_and(subset, dataset['SCHISM_hgrid_node_x'] > bounds[0])
            if bounds[2] is not None:
                subset = numpy.logical_and(subset, dataset['SCHISM_hgrid_node_x'] < bounds[2])
            if bounds[1] is not None:
                subset = numpy.logical_and(subset, dataset['SCHISM_hgrid_node_y'] > bounds[1])
            if bounds[3] is not None:
                subset = numpy.logical_and(subset, dataset['SCHISM_hgrid_node_y'] < bounds[3])

        return subset


class ElevationStationOutput(StationTimeSeriesOutput):
    """
    ``staout_1`` - Elevation Time Series at Specified Elevation Recording Stations

    """

    out_filename = 'schism_point_elevtion.nc'
    file_patterns = ['staout_1']
    variables = ['elevation']


class VelocityStationOutput(StationTimeSeriesOutput):
    """
    ``staout_7`` & ``staout_8`` - Depth-averaged Velocity Time Series
    at Specified Velocity Recording Stations

    """

    out_filename = 'schism_point_velocity.nc'
    file_patterns = ['staout_7', 'staout_8']
    variables = ['horizontalVelX', 'horizontalVelY']


class FieldOutput(SchismOutput, ABC):
    @classmethod
    def read(
        cls, filenames: Union[PathLike, List[PathLike]], names: List[str] = None
    ) -> Union[DataFrame, DataArray]:
        """
        Parse SCHISM output files

        :param filenames: file path to SCHISM NetCDF outputs
        :param names: list of data variables to extract
        :return: parsed data
        """

        if not isinstance(filenames, List):
            filenames = [filenames]

        if all(any(j.matches(i) for j in filenames) for i in cls.file_patterns):
            raise ValueError(
                f'{cls.__name__} requires files that match {" & ".join(cls.file_patterns)}'
                f' input files for {" & ".join(cls.variables)}'
            )

        filenames = [Path(fnm) for fnm in filenames]

        for filename in filenames:
            LOGGER.debug(f'opening "{"/".join(filename.parts[-2:])}"')

        if names is None:
            names = []
            for subclass in FieldOutput.__subclasses__():
                # NOTE: Same filename can contain info for multiple vars (out2d)
                if all(
                    [
                        fpath.name == Path(predef).name
                        for fpath, predef in zip(filenames, subclass.filenames)
                    ]
                ):
                    names.extend(subclass.variables)
            else:
                raise NotImplementedError(
                    f'Support for one of the provided SCHISM output files is not implemented'
                )
        names = list(set(names))

        dataset = xarray.open_mfdataset(filenames, drop_variables=cls.drop_variables)
        data = dataset[names]
        data = data.assign_attrs(**dataset.attrs)

        for filename in filenames:
            LOGGER.debug(f'finished reading "{"/".join(filename.parts[-2:])}"')

        return data

    @classmethod
    def read_directory(
        cls, directory: PathLike, variables: List[str] = None, parallel: bool = False
    ) -> Dataset:
        """
        Compile a dataset from output files in the given directory.

        :param directory: directory containing output files
        :param variables: variables to return
        :param parallel: load data concurrently with Dask
        :return: dataset of output data
        """

        if not isinstance(directory, Path):
            directory = Path(directory)

        if variables is None:
            variables = cls.variables

        output_dict = find_run_dir_for_output(cls.file_patterns, directory)

        file_collection = [
            [[f for f in patt_dict['files']] for patt_dict in run_dict['outputs'].values()]
            for run_dict in output_dict.values()
        ]
        run_dirs = list(output_dict.keys())

        # Open only the first set of outputs to get all drop_variables
        drop_variables = deepcopy(cls.drop_variables)

        with xarray.open_mfdataset(
            [flist[0] for flist in file_collection[0]], drop_variables=drop_variables
        ) as sample_dataset:
            drop_variables.extend(
                variable_name
                for variable_name in sample_dataset.variables
                if variable_name not in variables
                and variable_name
                not in [
                    'time',
                    'SCHISM_hgrid_node_x',
                    'SCHISM_hgrid_node_y',
                    'depth',
                    'dryFlagNode',
                    'minimum_depth',
                ]
            )
            if not all(var in sample_dataset.data_vars for var in variables):
                LOGGER.warn("Files don't contain all the required variables!")
                return xarray.Dataset()

        # Now open all the relevant output files for all runs
        with dask.config.set(**{'array.slicing.split_large_chunks': True}):
            dataset = xarray.open_mfdataset(
                [f for flist in file_collection[0] for f in flist],
                drop_variables=drop_variables,
                parallel=parallel,
                lock=False,
            ).expand_dims({'run': [run_dirs[0].name]})
            for run_idx, run_dir in enumerate(run_dirs[1:]):
                run_out_filelist = [f for flist in file_collection[run_idx + 1] for f in flist]
                dataset = xarray.combine_nested(
                    [
                        dataset,
                        xarray.open_mfdataset(
                            run_out_filelist,
                            drop_variables=drop_variables,
                            parallel=parallel,
                            lock=False,
                        ).expand_dims({'run': [run_dir.name]}),
                    ],
                    concat_dim='run',
                )

        # Drop run dimension for variables fixed across runs
        # `SCHISM_hgrid_node_x` and `SCHISM_hgrid_node_y` are
        # coordinates and are not expanded in `run` dimension
        fixed_vars = [
            'node',
            'depth',
        ]
        for var in fixed_vars:
            if var not in dataset:
                continue
            dataset = dataset.assign({var: dataset[var].isel(run=0)})

        # Add element table
        dataset = cls._add_element_table(dataset, directory)

        # TODO: What if it's different for different runs?
        if 'minimum_depth' in dataset:
            dataset = dataset.assign_attrs(
                minimum_depth=dataset.minimum_depth.isel(one=0, run=0).values
            )

        # TODO: Does it make sense to have all these as "coord" so that
        # we can return a DataArray?!
        coord_vars = [
            'SCHISM_hgrid_node_x',
            'SCHISM_hgrid_node_y',
            'depth',
            'element',
            'node',
            'dryFlagNode',
        ]
        for var in coord_vars:
            if var in dataset:
                dataset = dataset.assign_coords({var: dataset[var]})

        ret_value = dataset[variables]
        if 'element' in dataset and 'element' not in ret_value:
            ret_value = ret_value.assign_coords({'element': dataset['element']})
        ret_value = ret_value.assign_attrs(**dataset.attrs)
        return ret_value

    @classmethod
    def subset(
        cls,
        dataset: Union[Dataset, DataArray],
        bounds: (float, float, float, float) = None,
        wind_swath: [str, int] = None,
        maximum_depth: float = None,
        minimum_depth: float = None,
        **kwargs,
    ) -> Union[Dataset, DataArray]:
        # TODO: Is this right?
        subset = ~dataset['nSCHISM_hgrid_node'].isnull()

        if wind_swath is not None:
            cyclone = wind_swath[0]
            isotach = wind_swath[1]
            LOGGER.debug(f'filtering within {cyclone} wind swath {isotach}')
            if not isinstance(cyclone, VortexTrack):
                try:
                    cyclone = VortexTrack.from_file(cyclone)
                except FileNotFoundError:
                    cyclone = VortexTrack(cyclone)
            swath = cyclone.wind_swaths(wind_speed=isotach)
            if 'BEST' in swath:
                tracks = swath['BEST']
            elif 'OFCL' in swath:
                tracks = swath['OFCL']
            else:
                raise ValueError(
                    'Neither best or official track could be found for the specified storm'
                )

            series = pandas.Series(tracks.keys())
            latest_track = tracks[series[pandas.to_datetime(series).argmax()]]

            polygon = GeoDataFrame(index=[0], geometry=[latest_track])
            with dask.config.set(**{'array.slicing.split_large_chunks': True}):
                geometry = geopandas.points_from_xy(
                    dataset['SCHISM_hgrid_node_x'].values,
                    dataset['SCHISM_hgrid_node_y'].values,
                )
                points = GeoDataFrame(
                    {
                        'lon': dataset['SCHISM_hgrid_node_x'].values,
                        'lat': dataset['SCHISM_hgrid_node_y'].values,
                    },
                    geometry=geometry,
                )
                inpoly = geopandas.tools.sjoin(points, polygon, predicate='within', how='left')
            subset = numpy.logical_and(subset, ~numpy.isnan(inpoly.index_right.values))

        if bounds is not None:
            LOGGER.debug(f'filtering within bounds {bounds}')
            if bounds[0] is not None:
                subset = numpy.logical_and(subset, dataset['SCHISM_hgrid_node_x'] > bounds[0])
            if bounds[2] is not None:
                subset = numpy.logical_and(subset, dataset['SCHISM_hgrid_node_x'] < bounds[2])
            if bounds[1] is not None:
                subset = numpy.logical_and(subset, dataset['SCHISM_hgrid_node_y'] > bounds[1])
            if bounds[3] is not None:
                subset = numpy.logical_and(subset, dataset['SCHISM_hgrid_node_y'] < bounds[3])

        if maximum_depth is not None:
            LOGGER.debug(f'filtering by maximum depth {maximum_depth}')
            subset = numpy.logical_and(subset, dataset['depth'] <= maximum_depth)

        if minimum_depth is not None:
            LOGGER.debug(f'filtering by minimum depth {minimum_depth}')
            subset = numpy.logical_and(subset, dataset['depth'] >= minimum_depth)

        return subset

    @classmethod
    def _add_element_table(cls, dataset: Dataset, directory: PathLike) -> Dataset:
        # hgrid.gr3 or ll
        gridfile_pattern = 'hgrid.*'
        matches = list(directory.glob(f'**/{gridfile_pattern}'))

        if len(matches) == 0:
            return dataset

        # TODO: Check if all the found hgrid files are the same
        gridfile = matches[0]

        # NOTE: All elements are treated as tria (quads are split)
        grid = Hgrid.open(gridfile, crs=4326)
        dataset = dataset.assign(
            element=xarray.DataArray(
                data=grid.elements.triangulation.triangles, dims=('nele', 'nvertex')
            )
        )

        return dataset


class ExtremumScalarFieldOutputCalculator(FieldOutput):

    derived_name: str
    derived_time_name: str
    extermum_func: str

    @classmethod
    def read(
        cls, filenames: Union[PathLike, List[PathLike]], names: List[str] = None
    ) -> Union[DataFrame, DataArray]:
        full_ds = super().read(filenames, names)
        ds = cls._calc_extermum(full_ds)
        return ds

    @classmethod
    def read_directory(
        cls, directory: PathLike, variables: List[str] = None, parallel: bool = False
    ) -> Dataset:
        if variables is None:
            variables = cls.variables
        full_ds = super().read_directory(directory, variables, parallel)
        if all(var in full_ds.data_vars for var in variables):
            ds = cls._calc_extermum(full_ds)
            return ds

        return xarray.Dataset()

    @classmethod
    def _calc_extermum(cls, full_ds) -> Dataset:
        if len(cls.variables) > 1:
            to_extrm_ary = (
                numpy.sum(getattr(full_ds, var) ** 2 for var in cls.variables) ** 0.5
            )
        else:
            to_extrm_ary = full_ds[cls.variables[0]]

        uses_dask_array = False
        if to_extrm_ary.chunks is not None and len(to_extrm_ary.chunks) > 0:
            uses_dask_array = True

        if uses_dask_array:
            # compute() due to https://github.com/pydata/xarray/issues/2511
            arg_extrm_var = getattr(to_extrm_ary, cls.extermum_func)(dim='time').compute()
        else:
            arg_extrm_var = getattr(to_extrm_ary, cls.extermum_func)(dim='time')
        extrm_vals = to_extrm_ary.isel(time=arg_extrm_var)
        extrm_times = to_extrm_ary.time.isel(time=arg_extrm_var)

        # TODO: Chunk the dataset?
        ds = xarray.Dataset(
            {
                cls.derived_name: (extrm_vals.dims, extrm_vals.data),
                cls.derived_time_name: (extrm_vals.dims, extrm_times.data),
                'run': extrm_vals.run.data,
                'nSCHISM_hgrid_node': extrm_vals.nSCHISM_hgrid_node.data,
            },
        )
        ds = ds.assign_attrs(**full_ds.attrs)
        if 'SCHISM_hgrid_node_x' in full_ds.data_vars:
            ds['SCHISM_hgrid_node_x'] = full_ds.SCHISM_hgrid_node_x
        elif 'SCHISM_hgrid_node_x' in full_ds.coords:
            ds = ds.assign_coords({'SCHISM_hgrid_node_x': full_ds.SCHISM_hgrid_node_x})
        if 'SCHISM_hgrid_node_y' in full_ds.data_vars:
            ds['SCHISM_hgrid_node_y'] = full_ds.SCHISM_hgrid_node_y
        elif 'SCHISM_hgrid_node_y' in full_ds.coords:
            ds = ds.assign_coords({'SCHISM_hgrid_node_y': full_ds.SCHISM_hgrid_node_y})

        return ds


class MaximumScalarFieldOutputCalculator(ExtremumScalarFieldOutputCalculator):
    extermum_func = 'argmax'


class MinimumScalarFieldOutputCalculator(ExtremumScalarFieldOutputCalculator):
    extermum_func = 'argmin'


class MaximumElevationOutput(MaximumScalarFieldOutputCalculator):
    """
    ``out2d.nc`` - Derived Maximum Elevation at All Nodes in the Model Grid

    """

    out_filename = 'schism_max_elevation.nc'
    file_patterns = ['out2d_*.nc']
    variables = ['elevation']
    derived_name = 'max_elevation'
    derived_time_name = 'max_elevation_times'

    @classmethod
    def read(
        cls, filenames: Union[PathLike, List[PathLike]], names: List[str] = None
    ) -> Union[DataFrame, DataArray]:
        dataset = super().read(filenames, names)
        dataset.attrs['h0'] = dataset.minimum_depth
        return cls._set_dry_to_null(dataset)

    @classmethod
    def read_directory(
        cls, directory: PathLike, variables: List[str] = None, parallel: bool = False
    ) -> Dataset:

        dataset = super().read_directory(directory, variables, parallel)
        dataset.attrs['h0'] = dataset.minimum_depth
        return cls._set_dry_to_null(dataset)

    @classmethod
    def _set_dry_to_null(cls, dataset: Dataset) -> Dataset:
        dataset[cls.derived_name] = dataset[cls.derived_name].where(
            dataset[cls.derived_name] > dataset.attrs['h0'], numpy.nan
        )

        return dataset


class MaximumVelocityOutput(MaximumScalarFieldOutputCalculator):
    """
    ``horizontalVelX.nc`` and ``horizontalVelY.nc`` - Derived Maximum Speed at All Nodes in the Model Grid

    """

    out_filename = 'schism_max_velocity.nc'
    file_patterns = ['horizontalVelX_*.nc', 'horizontalVelY_*.nc']
    variables = ['horizontalVelX', 'horizontalVelY']
    derived_name = 'max_velocity'
    derived_time_name = 'max_velocity_times'


class MinimumSurfacePressureOutput(MinimumScalarFieldOutputCalculator):
    """
    ``out2d.nc`` - Minimum Sea-level Pressure at All Nodes in the Model Grid

    """

    out_filename = 'schism_min_pressure.nc'
    file_patterns = ['out2d_*.nc']
    variables = ['airPressure']
    derived_name = 'min_pressure'
    derived_time_name = 'min_pressure_times'


class MaximumSurfaceWindOutput(MaximumScalarFieldOutputCalculator):
    """
    ``out2d.nc`` - Maximum Surface Wind Speed at All Nodes in the Model Grid

    """

    out_filename = 'schism_max_wind.nc'
    file_patterns = ['out2d_*.nc']
    variables = ['windSpeedX', 'windSpeedY']
    derived_name = 'max_wind'
    derived_time_name = 'max_wind_times'


class FieldTimeSeriesOutput(FieldOutput, TimeSeriesOutput, ABC):
    pass


class ElevationTimeSeriesOutput(FieldTimeSeriesOutput):
    """
    ``out2d.nc`` - Elevation Time Series at All Nodes in the Model Grid

    """

    out_filename = 'schism_elevation.nc'
    file_patterns = ['out2d_*.nc']
    variables = ['elevation']

    @classmethod
    def read(
        cls, filenames: Union[PathLike, List[PathLike]], names: List[str] = None
    ) -> Union[DataFrame, DataArray]:
        dataset = super().read(filenames, names)
        dataset.attrs['h0'] = dataset.minimum_depth
        return cls._set_dry_to_null(dataset)

    @classmethod
    def read_directory(
        cls, directory: PathLike, variables: List[str] = None, parallel: bool = False
    ) -> Dataset:

        dataset = super().read_directory(directory, variables, parallel)
        dataset.attrs['h0'] = dataset.minimum_depth
        return cls._set_dry_to_null(dataset)

    @classmethod
    def _set_dry_to_null(cls, dataset: Dataset) -> Dataset:
        for var in cls.variables:
            dataset[var] = dataset[var].where(dataset['dryFlagNode'] == 0, numpy.nan)

        return dataset

    @classmethod
    def subset(
        cls,
        dataset: Union[Dataset, DataArray],
        bounds: (float, float, float, float) = None,
        maximum_depth: float = None,
        elevation_selection: ElevationSelection = None,
        **kwargs,
    ) -> Union[Dataset, DataArray]:
        subset = super().subset(dataset, bounds=bounds, maximum_depth=maximum_depth)

        if elevation_selection is not None:
            if not isinstance(elevation_selection, ElevationSelection):
                elevation_selection = convert_value(elevation_selection, ElevationSelection)

            dry_subset = dataset['dryFlagNode'] == 1

            if elevation_selection == ElevationSelection.wet:
                elevation_subset = ~dry_subset.any('time')
            elif elevation_selection == ElevationSelection.inundated:
                # get all nodes that experienced inundation (were both wet and dry at any time)
                elevation_subset = dry_subset.any('time') & ~dry_subset.all('time')
            else:
                elevation_subset = dry_subset.all('time')

            if 'run' in dataset:
                elevation_subset = elevation_subset.any('run')

            subset = numpy.logical_and(subset, elevation_subset)

        return subset


class VelocityTimeSeriesOutput(FieldTimeSeriesOutput):
    """
    ``horizontalVelX.nc`` and ``horizontalVelY.nc`` - Depth-averaged Velocity Time Series at All Nodes in the Model Grid

    """

    out_filename = 'schism_velocity.nc'
    file_patterns = ['horizontalVelX_*.nc', 'horizontalVelY_*.nc']
    variables = ['horizontalVelX', 'horizontalVelY']


class SurfacePressureTimeSeriesOutput(FieldTimeSeriesOutput):
    """
    ``out2d.nc`` - Sea-level Pressure Time Series at All Nodes in the Model Grid

    """

    out_filename = 'schism_pressure.nc'
    file_patterns = ['out2d_*.nc']
    variables = ['airPressure']


class SurfaceWindTimeSeriesOutput(FieldTimeSeriesOutput):
    """
    ``out2d.nc`` - Surface Wind Velocity Time Series at All Nodes in the Model Grid

    """

    out_filename = 'schism_wind.nc'
    file_patterns = ['out2d_*.nc']
    variables = ['windSpeedX', 'windSpeedY']


class _GlobDict(UserDict):
    """A dictionary that tries to match key by unix glob if the key is not found"""

    def __getitem__(self, query_key):
        if query_key in self.data:
            return self.data[query_key]
        for stored_key in self.data:
            if fnmatch.fnmatch(query_key, stored_key):
                return self.data[stored_key]

        return super().__getitem__(query_key)


def schism_file_data_variables(cls: type = None, existing_dict=None) -> Dict[str, List[str]]:

    file_data_variables = _GlobDict()
    if existing_dict is not None:
        file_data_variables = deepcopy(existing_dict)

    if cls is None:
        cls = SchismOutput

    for subclass in cls.__subclasses__():
        try:
            for patt in subclass.file_patterns:
                file_data_variables.setdefault(patt, set()).add(subclass)
        except AttributeError:
            file_data_variables = schism_file_data_variables(subclass, file_data_variables)
    return file_data_variables


SCHISM_FILE_OUTPUTS = schism_file_data_variables()


def parse_schism_outputs(
    directory: PathLike = None, file_outputs: List[str] = None, parallel: bool = False,
) -> Dict[str, dict]:
    """
    Parse output from multiple SCHISM runs.

    :param directory: directory containing run output directories
    :param file_outputs: output files to parse
    :param parallel: load data concurrently with Dask
    :return: variables to parsed data
    """

    if directory is None:
        directory = Path.cwd()
    elif not isinstance(directory, Path):
        directory = Path(directory)

    if file_outputs is None:
        file_outputs = SCHISM_FILE_OUTPUTS
    elif isinstance(file_outputs, Collection):
        file_outputs = {pattern: SCHISM_FILE_OUTPUTS[pattern] for pattern in file_outputs}
    elif isinstance(file_outputs, Mapping):
        file_outputs = {
            pattern: subclasses if subclasses is not None else SCHISM_FILE_OUTPUTS[pattern]
            for pattern, subclasses in file_outputs.items()
        }

    output_tree = {}
    node_info_keys = [
        'SCHISM_hgrid_node_x',
        'SCHISM_hgrid_node_y',
        'depth',
        'dryFlagNode',
        'element',
    ]
    node_info_data = xarray.Dataset()
    for basename, output_classes in file_outputs.items():
        for output_class in output_classes:
            try:
                # Some classes match multiple patterns (i.e. basename)
                if output_class in output_tree:
                    continue

                dataset = output_class.read_directory(
                    directory, variables=output_class.variables, parallel=parallel,
                )
                if len(dataset) == 0:
                    continue
                if all(
                    var in dataset.data_vars or var in dataset.coords for var in node_info_keys
                ):
                    node_info_data = dataset[node_info_keys]

                # NOTE: The dataset variable might be derived variables
                # skip_ds = False
                # for var in output_class.variables:
                #     if var in dataset.data_vars:
                #         continue
                #     skip_ds = True
                # if skip_ds:
                #     continue

                output_tree[output_class] = dataset

            except (ValueError, FileNotFoundError) as error:
                LOGGER.warning(error)

    if len(node_info_data.coords) != 0:
        for output_class, dataset in output_tree.items():
            if 'nSCHISM_hgrid_node' not in dataset.dims:
                continue
            output_tree[output_class] = dataset.merge(node_info_data)
    else:
        LOGGER.warn('No dataset found with infomration about node locations and dry nodes!')

    return output_tree


def combine_outputs(
    directory: PathLike = None,
    file_data_variables: Dict[str, List[str]] = None,
    bounds: Tuple[float, float, float, float] = None,
    maximum_depth: float = None,
    elevation_selection: ElevationSelection = None,
    output_directory: PathLike = None,
    parallel: bool = False,
) -> Dict[str, DataFrame]:
    if directory is None:
        directory = Path.cwd()
    elif not isinstance(directory, Path):
        directory = Path(directory)

    if maximum_depth is not None and not isinstance(maximum_depth, float):
        maximum_depth = float(maximum_depth)

    if output_directory is not None:
        if not isinstance(output_directory, Path):
            output_directory = Path(output_directory)
        if not output_directory.exists():
            output_directory.mkdir(parents=True, exist_ok=True)

    runs_directory = directory / 'runs'
    if not runs_directory.exists():
        raise FileNotFoundError(f'runs directory does not exist at "{runs_directory}"')

    track_directory = directory / 'track_files'
    if not track_directory.exists():
        raise FileNotFoundError(f'track directory does not exist at "{track_directory}"')

    if file_data_variables is None:
        file_data_variables = SCHISM_FILE_OUTPUTS
    elif isinstance(file_data_variables, Collection):
        file_data_variables = {
            filename: SCHISM_FILE_OUTPUTS[filename] for filename in file_data_variables
        }
    elif isinstance(file_data_variables, Mapping):
        file_data_variables = {
            filename: variables if variables is not None else SCHISM_FILE_OUTPUTS[filename]
            for filename, variables in file_data_variables.items()
        }

    # parse all the inputs using built-in parser
    output_data = {'perturbations.nc': parse_vortex_perturbations(track_directory)}

    # parse all the outputs using built-in parser
    LOGGER.info(f'parsing from "{directory}"')
    parsed_files = parse_schism_outputs(
        directory=runs_directory, file_outputs=file_data_variables, parallel=parallel,
    )

    wetdry_timeseries_output = ElevationTimeSeriesOutput
    if elevation_selection is not None and wetdry_timeseries_output not in parsed_files:
        raise ValueError(f'elevation time series not found')

    output_data.update({k.out_filename: v for k, v in parsed_files.items()})

    # generate subset
    elevation_subset = None
    for output_class, file_data in parsed_files.items():
        if 'nSCHISM_hgrid_node' in file_data:
            num_nodes = len(file_data['nSCHISM_hgrid_node'])

            variable_shape_string = ', '.join(
                f'"{name}" {variable.shape}' for name, variable in file_data.items()
            )
            LOGGER.info(
                f'found {len(file_data)} variable(s) in "{output_class.file_patterns}": {variable_shape_string}'
            )

            subset = ~file_data['nSCHISM_hgrid_node'].isnull()

            if elevation_subset is not None:
                subset = numpy.logical_and(subset, elevation_subset)

            subset = numpy.logical_and(
                subset,
                output_class.subset(
                    file_data,
                    bounds=bounds,
                    maximum_depth=maximum_depth,
                    elevation_selection=elevation_selection,
                ),
            )

            if subset is not None:
                file_data = file_data.sel(nSCHISM_hgrid_node=subset)

                LOGGER.info(
                    f'subsetted {len(file_data["nSCHISM_hgrid_node"])} out of {num_nodes} total nodes ({len(file_data["nSCHISM_hgrid_node"]) / num_nodes:3.2%})'
                )

                if elevation_selection is not None:
                    elevation_subset = subset

            output_data[output_class.out_filename] = file_data

    if output_directory is not None:
        for key, file_data in output_data.items():
            basename = key
            if isinstance(key, type) and issubclass(key, SchismOutput):
                basename = key.out_filename
            output_filename = output_directory / basename
            LOGGER.info(f'writing to "{output_filename}"')
            file_data.to_netcdf(
                output_filename,
                # encoding={
                #     variable_name: {'zlib': True} for variable_name in file_data.variables
                # },
            )

    return output_data


def subset_dataset(
    ds: Dataset,
    variable: str,
    maximum_depth: float = None,
    minimum_depth: float = None,
    wind_swath: list = None,
    bounds: (float, float, float, float) = None,
    node_status_selection: dict = None,
    point_spacing: int = None,
    output_filename: PathLike = None,
):
    with dask.config.set(**{'array.slicing.split_large_chunks': True}):
        if node_status_selection is None:
            node_status_mask = None
        elif node_status_selection['mask'] == 'sometimes_wet':
            node_status_mask = (
                ~ds[variable].sel(run=node_status_selection['runs']).isnull().all('run'),
            )
        elif node_status_selection['mask'] == 'always_wet':
            node_status_mask = (
                ~ds[variable].sel(run=node_status_selection['runs']).isnull().any('run'),
            )
        else:
            raise ValueError(
                f'node_status_selection {node_status_selection["mask"]} unrecognized'
            )

        node_subset_mask = (
            FieldOutput.subset(
                ds['nSCHISM_hgrid_node'],
                maximum_depth=maximum_depth,
                minimum_depth=minimum_depth,
                bounds=bounds,
                wind_swath=wind_swath,
            ),
        )
        if node_status_mask is None:
            node_status_mask = node_subset_mask
        subsetted_nodes = ds['nSCHISM_hgrid_node'].values[
            numpy.logical_and(node_status_mask, node_subset_mask).squeeze()
        ]
        if point_spacing is not None:
            subsetted_nodes = subsetted_nodes[::point_spacing]
        subset = ds.sel(nSCHISM_hgrid_node=subsetted_nodes)

        # TODO: Used to be code to adjust element table if present based on node

        try:
            subset = subset.drop_sel(run='original')
        except:
            pass
        if len(subset['nSCHISM_hgrid_node']) != len(ds['nSCHISM_hgrid_node']):
            LOGGER.info(
                f'subsetted down to {len(subset["nSCHISM_hgrid_node"])} nodes ({len(subset["nSCHISM_hgrid_node"]) / len(ds["nSCHISM_hgrid_node"]):.1%})'
            )
        if output_filename is not None:
            if not isinstance(output_filename, Path):
                output_filename = Path(output_filename)
            LOGGER.info(f'saving subset to "{output_filename}"')
            subset.to_netcdf(output_filename)

    return subset


def extrapolate_water_elevation_to_dry_areas(
    da: DataArray,
    k_neighbors: int = 1,
    idw_order: int = 1,
    compute_headloss: bool = False,
    mann_coef: float = 0.05,
    u_ref: float = 0.4,
    d_ref: float = 1.0,
):
    # return a deep copy of original datarrary on output
    da_adjusted = da.copy(deep=True)

    # Get coordinates in conformal projection (e.g,, Mercator)
    # for determining closest distance
    crs_from = 'EPSG:4326'  # WGS84
    crs_to = 'EPSG:3857'  # Mercator
    transformer = Transformer.from_crs(crs_from=crs_from, crs_to=crs_to, always_xy=True)
    x, y = transformer.transform(da['x'].values, da['y'].values)
    projected_coordinates = numpy.vstack([x, y]).T

    # for mapping back to node numbers
    nodes = numpy.arange(da.sizes['nSCHISM_hgrid_node'])

    # compute the friction factor for headloss calculation:
    # Rucker, et al. (2021). Natural Hazards.
    # https://doi.org/10.1007/s11069-021-04634-8
    if compute_headloss:
        friction_factor = (u_ref * mann_coef) ** 2 / d_ref ** (4 / 3)
    else:
        friction_factor = 0.0

    # inverse distance weighting of order `idw_order` with `k_nearest` neighbors
    for run in range(da.sizes['run']):
        null = numpy.isnan(da[run, :])
        tree = KDTree(projected_coordinates[~null])
        dd, nn = tree.query(projected_coordinates[null], k=k_neighbors)
        if k_neighbors == 1:
            headloss = dd * friction_factor  # hydraulic friction loss
            da_adjusted[run, null] = da[run, nodes[~null][nn]].values - headloss
        else:
            for kk in range(k_neighbors):
                weights = dd[:, kk] ** (-idw_order)
                headloss = dd[:, kk] * friction_factor  # hydraulic friction loss
                total_head = da[run, nodes[~null][nn[:, kk]]].values - headloss
                if kk == 0:
                    idw_sum = total_head * weights
                    weight_sum = weights
                else:
                    idw_sum += total_head * weights
                    weight_sum += weights
            da_adjusted[run, null] = idw_sum / weight_sum

    return da_adjusted


def convert_schism_output_dataset_to_adcirc_like(schism_ds: Dataset) -> Dataset:

    # TODO: Add sanity check for schism dataset

    coord_vars = []
    if 'station_index' in schism_ds.data_vars:
        # Station data
        temp_ds = schism_ds.swap_dims({'station_index': 'station'}).reset_coords(
            'station_index'
        )
        coord_vars.append('time')
    else:
        # Field data
        temp_ds = schism_ds.copy()
        if 'time' in temp_ds.data_vars:
            coord_vars.append('time')
        if 'SCHISM_hgrid_node_x' in temp_ds.data_vars:
            coord_vars.append('x')
        if 'SCHISM_hgrid_node_y' in temp_ds.data_vars:
            coord_vars.append('y')

    temp_ds = temp_ds.rename(
        **{k: v for k, v in SCHISM_ADCIRC_VAR_MAPPING.items() if k in schism_ds.data_vars}
    )
    temp_ds = temp_ds.rename(
        **{k: v for k, v in SCHISM_ADCIRC_COORD_MAPPING.items() if k in schism_ds.coords}
    )

    if 'depth' in temp_ds:
        temp_ds = temp_ds.assign_coords({'depth': temp_ds['depth']})

    if 'element' in temp_ds:
        temp_ds = temp_ds.assign_coords({'element': temp_ds['element']})

    if 'node' in temp_ds:
        temp_ds = temp_ds.assign_coords({'node': temp_ds['node']})

    if 'run' in temp_ds.coords:
        coord_vars.insert(0, 'run')

    adcirc_ds = temp_ds.set_coords(coord_vars)

    return adcirc_ds


def convert_schism_output_files_to_adcirc_like(
    directory: PathLike = None,
    file_data_variables: Dict[str, List[str]] = None,
    bounds: Tuple[float, float, float, float] = None,
    maximum_depth: float = None,
    elevation_selection: ElevationSelection = None,
    output_directory: PathLike = None,
    parallel: bool = False,
) -> None:

    if output_directory is not None:
        if not isinstance(output_directory, Path):
            output_directory = Path(output_directory)
        if not output_directory.exists():
            output_directory.mkdir(parents=True, exist_ok=True)

    # NOTE: Don't pass the output_directory here
    results = combine_outputs(
        directory=directory,
        file_data_variables=file_data_variables,
        bounds=bounds,
        maximum_depth=maximum_depth,
        elevation_selection=elevation_selection,
        output_directory=None,
        parallel=parallel,
    )

    output_data = {}
    for out_name, data in results.items():
        newkey = SCHISM_ADCIRC_OUT_MAPPING.get(out_name, None)
        if newkey is None:
            newkey = out_name

        if newkey != 'perturbations.nc':
            data = convert_schism_output_dataset_to_adcirc_like(data)
        output_data[newkey] = data

    if output_directory is not None:
        for out_name, file_data in output_data.items():
            output_filename = output_directory / out_name
            LOGGER.info(f'writing to "{output_filename}"')
            file_data.to_netcdf(
                output_filename,
                # encoding={
                #     variable_name: {'zlib': True} for variable_name in file_data.variables
                # },
            )

    return output_data



from os import PathLike
from pathlib import Path
from typing import List, Union

import chaospy
import numpoly
import numpy
import sklearn
import xarray

from ensembleperturbation.utilities import get_logger

LOGGER = get_logger('surrogate')


def surrogate_from_karhunen_loeve(
    mean_vector: numpy.ndarray,
    eigenvalues: numpy.ndarray,
    modes: numpy.ndarray,
    kl_surrogate_model: numpoly.ndpoly,
    filename: PathLike = None,
) -> numpoly.ndpoly:
    """
    build a polynomial from the given Karhunen-Loeve expansion (eigenvalues and modes) along with a mean vector along the node space

    the joint Karhunen-Loeve / Polynomial Chaos polynomial is derived from Eq (6) in Sargsyan, K. and Ricciuto D. (2021):

    .. math::

        \sum_{k=0}^{K} ( \delta_{k0} * \overline{f(z)} + \sum_{j=0}^{L} [ c_{jk} * \sqrt{ev_j} * ef_j(z) ] )

    ``K`` -> num PC coefficients

    ``L`` -> num KL modes

    ``c`` -> PC coefficient

    ``ev`` -> eigenvalue

    ``ef`` -> eigenfunction

    :param mean_vector: all points
    :param eigenvalues: eigenvalues of each point to each mode
    :param modes: modes of the Karhunen-Loeve expansion
    :param kl_surrogate_model: ``ndpoly`` surrogate polynomial generated on training set
    :param filename: file path to save surrogate as pickled ``ndpoly`` object
    :return: polynomial constructued from the above equation
    """

    if filename is not None and not isinstance(filename, Path):
        filename = Path(filename)

    num_points = len(mean_vector)
    num_modes = len(eigenvalues)

    assert num_modes == len(
        kl_surrogate_model
    ), 'number of kl_dict eigenvalues must be equal to the length of the kl_surrogate_model'

    LOGGER.info(f'transforming surrogate to {num_points} points from {num_modes} eigenmodes')
    if filename is None or not filename.exists():
        # get the coefficients of the PC for each point in z (spatiotemporal dimension)
        pc_exponents = kl_surrogate_model.exponents
        pc_coefficients = numpy.array(kl_surrogate_model.coefficients)
        klpc_coefficients = numpy.sum(
            numpy.stack(
                [
                    numpy.dot(
                        (pc_coefficients * numpy.sqrt(eigenvalues))[:, mode_index, None],
                        modes[None, mode_index, :],
                    )
                    for mode_index in range(num_modes)
                ],
                axis=0,
            ),
            axis=0,
        )
        klpc_coefficients[0, :] += mean_vector

        surrogate_model = numpoly.ndpoly.from_attributes(
            exponents=pc_exponents, coefficients=klpc_coefficients,
        )

        if filename is not None:
            with open(filename, 'wb') as surrogate_file:
                LOGGER.info(f'saving surrogate model to "{filename}"')
                surrogate_model.dump(surrogate_file)
    else:
        LOGGER.info(f'loading surrogate model from "{filename}"')
        surrogate_model = chaospy.load(filename, allow_pickle=True)

    return surrogate_model


def surrogate_from_samples(
    samples: xarray.DataArray,
    perturbations: xarray.DataArray,
    polynomials: numpoly.ndpoly,
    regression_model: sklearn.linear_model,
    norms: numpy.ndarray = None,
    quadrature: bool = False,
    quadrature_weights: xarray.DataArray = None,
) -> numpoly.ndpoly:
    # create surrogate models for selected nodes
    if quadrature:
        LOGGER.info(
            f'fitting polynomial surrogate to {samples.shape} samples along the quadrature'
        )
        if quadrature_weights is None:
            LOGGER.warning('no quadrature weights provided')
        try:
            surrogate_model = chaospy.fit_quadrature(
                orth=polynomials,
                nodes=perturbations.T,
                weights=quadrature_weights,
                solves=samples,
                norms=norms,
            )
        except AssertionError:
            if (
                perturbations.shape[0] != len(quadrature_weights)
                or len(quadrature_weights) != len(samples)
                or len(samples) != perturbations.shape[0]
            ):
                raise AssertionError(
                    f'{perturbations.shape[0]} != {len(quadrature_weights)} != {len(samples)}'
                )
            else:
                raise
    else:
        LOGGER.info(
            f'fitting polynomial surrogate to {samples.shape} samples using regression'
        )
        try:
            poly_list = [None] * samples.shape[1]
            for mode in range(samples.shape[1]):
                poly_list[mode] = chaospy.fit_regression(
                    polynomials=polynomials,
                    abscissas=perturbations.T,
                    evals=samples[:, mode],
                    model=regression_model,
                )
            surrogate_model = numpoly.polynomial(poly_list)
            ## Or just call this for default regression
            # surrogate_model = chaospy.fit_regression(polynomials=polynomials,abscissas=perturbations.T,evals=samples)
        except AssertionError:
            if perturbations.T.shape[-1] != len(samples):
                raise AssertionError(f'{perturbations.T.shape[-1]} != {len(samples)}')
            else:
                raise

    return surrogate_model


def surrogate_from_training_set(
    training_set: xarray.Dataset,
    training_perturbations: xarray.Dataset,
    distribution: chaospy.Distribution,
    filename: PathLike = None,
    use_quadrature: bool = False,
    polynomial_order: int = 3,
    regression_model: sklearn.linear_model = sklearn.linear_model.LinearRegression(
        fit_intercept=False
    ),
) -> numpoly.ndpoly:
    """
    use ``chaospy`` to build a surrogate model from the given training set / perturbations and single / joint distribution

    :param training_set: array of data along nodes in the mesh to use to fit the model
    :param training_perturbations: perturbations along each variable space that comprise the cloud of model inputs
    :param distribution: ``chaospy`` distribution
    :param filename: path to file to store polynomial
    :param use_quadrature: assume that the variable perturbations and training set are built along a quadrature, and fit accordingly
    :param polynomial_order: order of the polynomial chaos expansion
    :param regression_model: the scikit linear model to use to fit the polynomial regression [LinearRegression (OLS) by default]: https://scikit-learn.org/stable/modules/linear_model.html
    :return: polynomial
    """

    if filename is not None and not isinstance(filename, Path):
        filename = Path(filename)

    if filename is None or not filename.exists():
        # expand polynomials with polynomial chaos
        polynomial_expansion, norms = chaospy.generate_expansion(
            order=polynomial_order,
            dist=distribution,
            rule='three_terms_recurrence',
            retall=True,
        )

        surrogate_model = surrogate_from_samples(
            samples=training_set,
            perturbations=training_perturbations['perturbations'],
            polynomials=polynomial_expansion,
            norms=norms,
            quadrature=use_quadrature,
            quadrature_weights=training_perturbations['weights'] if use_quadrature else None,
            regression_model=regression_model,
        )

        if filename is not None:
            with open(filename, 'wb') as surrogate_file:
                LOGGER.info(f'saving surrogate model to "{filename}"')
                surrogate_model.dump(surrogate_file)
    else:
        LOGGER.info(f'loading surrogate model from "{filename}"')
        surrogate_model = chaospy.load(filename, allow_pickle=True)

    return surrogate_model


def sensitivities_from_surrogate(
    surrogate_model: numpoly.ndpoly,
    distribution: chaospy.Distribution,
    variables: [str],
    nodes: xarray.Dataset,
    element_table: xarray.DataArray = None,
    filename: PathLike = None,
) -> xarray.DataArray:
    """
    Get sensitivities of a given order for the surrogate model and distribution.

    :param surrogate_model: polynomial representing the surrogate model
    :param distribution: single or joint distribution of variable space
    :param variables: variable names
    :param nodes: dataset containing node information (nodes and XYZ coordinates) of mesh
    :param filename: filename to store sensitivities
    :return: array of sensitivities per node per variable
    """

    if filename is not None and not isinstance(filename, Path):
        filename = Path(filename)

    if filename is None or not filename.exists():
        LOGGER.info(f'extracting sensitivities from surrogate model and distribution')

        sensitivities = [
            chaospy.Sens_m(surrogate_model, distribution),
            chaospy.Sens_t(surrogate_model, distribution),
        ]

        sensitivities = numpy.stack(sensitivities)

        sensitivities = xarray.DataArray(
            sensitivities,
            coords={
                'order': ['main', 'total'],
                'variable': variables,
                'node': nodes['node'],
                'x': nodes['x'],
                'y': nodes['y'],
                'depth': nodes['depth'],
            },
            dims=('order', 'variable', 'node'),
        ).T

        sensitivities = sensitivities.to_dataset(name='sensitivities')

        if element_table is not None:
            sensitivities = sensitivities.assign_coords({'element': element_table})

        if filename is not None:
            LOGGER.info(f'saving sensitivities to "{filename}"')
            sensitivities.to_netcdf(filename)
    else:
        LOGGER.info(f'loading sensitivities from "{filename}"')
        sensitivities = xarray.open_dataset(filename)

    return sensitivities


def validations_from_surrogate(
    surrogate_model: numpoly.ndpoly,
    training_set: xarray.Dataset,
    training_perturbations: xarray.Dataset,
    validation_set: xarray.Dataset = None,
    validation_perturbations: xarray.Dataset = None,
    minimum_allowable_value: float = None,
    convert_from_log_scale: Union[bool, float] = False,
    convert_from_depths: Union[bool, float] = False,
    element_table: xarray.DataArray = None,
    filename: PathLike = None,
) -> xarray.Dataset:
    """


    :param surrogate_model: polynomial of surrogate model to query
    :param training_set: set of training data (across nodes and perturbations)
    :param training_perturbations: array of perturbations corresponding to training set
    :param validation_set: set of validation data (across nodes and perturbations)
    :param validation_perturbations: array of perturbations corresponding to validation set
    :param minimum_allowable_value: if surrogate prediction falls below this value set to NaN
    :param convert_from_log_scale: whether to take the exp() of the result
    :param convert_from_depths: whether to substract still water depth from the result
    :param filename: file path to which to save
    :return: array of validations
    """

    if filename is not None and not isinstance(filename, Path):
        filename = Path(filename)

    if filename is None or not filename.exists():
        LOGGER.info(f'running surrogate model on {training_set.shape} training samples')
        training_results = surrogate_model(*training_perturbations['perturbations'].T).T
        if isinstance(convert_from_log_scale, float):
            training_results = convert_from_log_scale ** training_results
        elif convert_from_log_scale:
            training_results = numpy.exp(training_results)
        if isinstance(convert_from_depths, (float, numpy.ndarray)):
            training_results -= convert_from_depths
        if minimum_allowable_value is not None:
            # compare to adjusted depths if provided
            too_small = training_results < minimum_allowable_value
            training_results[too_small] = numpy.nan
        if isinstance(convert_from_depths, (float, numpy.ndarray)) or convert_from_depths:
            training_results -= training_set['depth'].values

        training_results = numpy.stack([training_set, training_results], axis=0)
        training_results = xarray.DataArray(
            training_results,
            coords={'source': ['model', 'surrogate'], **training_set.coords},
            dims=('source', 'run', 'node'),
            name='training',
        )

        if validation_set is None:
            node_validation = training_results.to_dataset(name='results')
        else:
            LOGGER.info(
                f'running surrogate model on {validation_set.shape} validation samples'
            )
            node_validation = surrogate_model(*validation_perturbations['perturbations'].T).T
            if isinstance(convert_from_log_scale, float):
                node_validation = convert_from_log_scale ** node_validation
            elif convert_from_log_scale:
                node_validation = numpy.exp(node_validation)
            if isinstance(convert_from_depths, (float, numpy.ndarray)):
                node_validation -= convert_from_depths
            if minimum_allowable_value is not None:
                too_small = node_validation < minimum_allowable_value
                node_validation[too_small] = numpy.nan
            if isinstance(convert_from_depths, (float, numpy.ndarray)) or convert_from_depths:
                node_validation -= validation_set['depth'].values

            node_validation = numpy.stack([validation_set, node_validation], axis=0)
            node_validation = xarray.DataArray(
                node_validation,
                coords={'source': ['model', 'surrogate'], **validation_set.coords},
                dims=('source', 'run', 'node'),
                name='validation',
            )

            node_validation = xarray.combine_nested(
                [training_results.drop('type'), node_validation.drop('type')],
                concat_dim='type',
            )
            node_validation = node_validation.assign_coords(type=['training', 'validation'])
            node_validation = node_validation.to_dataset(name='results')

        if element_table is not None:
            node_validation = node_validation.assign_coords({'element': element_table})

        if filename is not None:
            LOGGER.info(f'saving validation to "{filename}"')
            node_validation.to_netcdf(filename)
    else:
        LOGGER.info(f'loading validation from "{filename}"')
        node_validation = xarray.open_dataset(filename)

    return node_validation


def statistics_from_surrogate(
    surrogate_model: numpoly.ndpoly,
    distribution: chaospy.Distribution,
    training_set: xarray.Dataset,
    filename: PathLike = None,
) -> xarray.Dataset:
    if filename is not None and not isinstance(filename, Path):
        filename = Path(filename)

    if filename is None or not filename.exists():
        LOGGER.info(
            f'gathering mean and standard deviation from surrogate on {training_set.shape} training samples'
        )
        surrogate_mean = chaospy.E(poly=surrogate_model, dist=distribution)
        surrogate_std = chaospy.Std(poly=surrogate_model, dist=distribution)
        modeled_mean = training_set.mean('run')
        modeled_std = training_set.std('run')

        surrogate_mean = xarray.DataArray(
            surrogate_mean, coords=modeled_mean.coords, dims=modeled_mean.dims,
        )
        surrogate_std = xarray.DataArray(
            surrogate_std, coords=modeled_std.coords, dims=modeled_std.dims,
        )

        node_statistics = xarray.Dataset(
            {
                'mean': xarray.combine_nested(
                    [surrogate_mean, modeled_mean], concat_dim='source'
                ).assign_coords({'source': ['surrogate', 'model']}),
                'std': xarray.combine_nested(
                    [surrogate_std, modeled_std], concat_dim='source'
                ).assign_coords({'source': ['surrogate', 'model']}),
                'difference': numpy.fabs(surrogate_std - modeled_std),
            }
        )

        if filename is not None:
            LOGGER.info(f'saving statistics to "{filename}"')
            node_statistics.to_netcdf(filename)
    else:
        LOGGER.info(f'loading statistics from "{filename}"')
        node_statistics = xarray.open_dataset(filename)

    return node_statistics


def percentiles_from_samples(
    samples: xarray.DataArray,
    percentiles: List[float],
    surrogate_model: numpoly.ndpoly,
    distribution: chaospy.Distribution,
    convert_from_log_scale: Union[bool, float] = False,
) -> xarray.DataArray:
    LOGGER.info(f'calculating {len(percentiles)} percentile(s): {percentiles}')
    # surrogate_percentiles = chaospy.Perc(
    #    poly=surrogate_model, q=percentiles, dist=distribution, sample=samples.shape[1],
    # )
    surrogate_percentiles = compute_surrogate_percentiles(
        poly=surrogate_model,
        q=percentiles,
        dist=distribution,
        convert_from_log_scale=convert_from_log_scale,
    )

    surrogate_percentiles = xarray.DataArray(
        surrogate_percentiles,
        coords={
            'quantile': percentiles,
            **{
                coord: values
                for coord, values in samples.coords.items()
                if coord not in ['run', 'type']
            },
        },
        dims=('quantile', *(dim for dim in samples.dims if dim not in ['run', 'type'])),
    )

    return surrogate_percentiles


def percentiles_from_surrogate(
    percentiles: List[float],
    surrogate_model: numpoly.ndpoly,
    distribution: chaospy.Distribution,
    training_set: xarray.Dataset,
    minimum_allowable_value: float = None,
    convert_from_log_scale: Union[bool, float] = False,
    convert_from_depths: Union[bool, float] = False,
    element_table: xarray.DataArray = None,
    filename: PathLike = None,
) -> xarray.Dataset:
    """


    :param percentiles: positions where percentiles are taken. Must be a number or a list, 
        where all values are on the interval ``[0, 100]``.
    :param surrogate_model: polynomial of surrogate model to query
    :parama distribution: surrogate model ``chaospy`` distribution model
    :param training_set: set of training data (across nodes and perturbations)
    :param minimum_allowable_value: if surrogate prediction falls below this value set to NaN
    :param convert_from_log_scale: whether to take the exp() of the result
    :param convert_from_depths: whether to substract still water depth from the result
    :param filename: file path to which to save
    :return: array of percentiles
    """

    if filename is not None and not isinstance(filename, Path):
        filename = Path(filename)

    if filename is None or not filename.exists():
        surrogate_percentiles = percentiles_from_samples(
            samples=training_set,
            percentiles=percentiles,
            surrogate_model=surrogate_model,
            distribution=distribution,
            convert_from_log_scale=convert_from_log_scale,
        )

        # before evaluating quantile for model set null water elevation to the ground elevation
        training_set = numpy.fmax(training_set, -training_set['depth'])
        modeled_percentiles = training_set.quantile(
            dim='run', q=surrogate_percentiles['quantile'] / 100
        )

        if isinstance(convert_from_depths, (float, numpy.ndarray)):
            surrogate_percentiles -= convert_from_depths
        if minimum_allowable_value is not None:
            too_small = (
                modeled_percentiles + training_set['depth']
            ).values < minimum_allowable_value
            modeled_percentiles.values[too_small] = numpy.nan
            too_small = surrogate_percentiles.values < minimum_allowable_value
            surrogate_percentiles.values[too_small] = numpy.nan
        if isinstance(convert_from_depths, (float, numpy.ndarray)) or convert_from_depths:
            surrogate_percentiles -= training_set['depth']

        modeled_percentiles.coords['quantile'] = surrogate_percentiles['quantile']

        node_percentiles = xarray.combine_nested(
            [surrogate_percentiles, modeled_percentiles], concat_dim='source'
        ).assign_coords(source=['surrogate', 'model'])

        node_percentiles = node_percentiles.to_dataset(name='quantiles')

        node_percentiles = node_percentiles.assign(
            differences=numpy.fabs(surrogate_percentiles - modeled_percentiles)
        )

        if element_table is not None:
            node_percentiles = node_percentiles.assign_coords({'element': element_table})

        if filename is not None:
            LOGGER.info(f'saving percentiles to "{filename}"')
            node_percentiles.to_netcdf(filename)
    else:
        LOGGER.info(f'loading percentiles from "{filename}"')
        node_percentiles = xarray.open_dataset(filename)

    return node_percentiles


def compute_surrogate_percentiles(
    poly: numpoly.ndpoly,
    q: List[float],
    dist: chaospy.Distribution,
    sample: int = 10000,
    convert_from_log_scale: Union[bool, float] = False,
    **kws,
):
    """
    Percentile function 
    *Modified to be able to convert from log scale

    Note that this function is an empirical function that operates using Monte
    Carlo sampling.

    Args:
        poly (numpoly.ndpoly):
            Polynomial of interest.
        q (numpy.ndarray):
            positions where percentiles are taken. Must be a number or an
            array, where all values are on the interval ``[0, 100]``.
        dist (Distribution):
            Defines the space where percentile is taken.
        sample (int):
            Number of samples used in estimation.
        convert_from_log_scale (bool):  
            Whether to take the exp() of the result

    Returns:
        (numpy.ndarray):
            Percentiles of ``poly`` with ``Q.shape=poly.shape+q.shape``.

    Examples:
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.polynomial([0.05*q0, 0.2*q1, 0.01*q0*q1])
        >>> chaospy.Perc(poly, [0, 5, 50, 95, 100], dist).round(2)
        array([[ 0.  , -3.29, -5.3 ],
               [ 0.  , -0.64, -0.04],
               [ 0.03, -0.01, -0.  ],
               [ 0.15,  0.66,  0.04],
               [ 1.61,  3.29,  5.3 ]])

    """
    poly = chaospy.aspolynomial(poly)
    shape = poly.shape
    poly = poly.ravel()

    q = numpy.asarray(q).ravel() / 100.0
    dim = len(dist)

    # Interior
    Z = dist.sample(sample, **kws).reshape(len(dist), sample)
    poly1 = poly(*Z)

    # Min/max
    ext = numpy.mgrid[(slice(0, 2, 1),) * dim].reshape(dim, 2 ** dim).T
    ext = numpy.where(ext, dist.lower, dist.upper).T
    poly2 = poly(*ext)
    poly2 = numpy.array([_ for _ in poly2.T if not numpy.any(numpy.isnan(_))]).T

    # Finish
    if poly2.shape:
        poly1 = numpy.concatenate([poly1, poly2], -1)
    if isinstance(convert_from_log_scale, float):
        poly1 = convert_from_log_scale ** poly1
    elif convert_from_log_scale:
        poly1 = numpy.exp(poly1)
    samples = poly1.shape[-1]
    poly1.sort()
    out = poly1.T[numpy.asarray(q * (samples - 1), dtype=int)]
    out = out.reshape(q.shape + shape)

    return out


def probability_field_from_samples(
    samples: xarray.Dataset,
    levels: List[float],
    surrogate_model: numpoly.ndpoly,
    distribution: chaospy.Distribution,
    minimum_allowable_value: float = None,
    convert_from_log_scale: Union[bool, float] = False,
    convert_from_depths: Union[bool, float] = False,
) -> xarray.DataArray:

    LOGGER.info(f'calculating {len(levels)} probability field(s): {levels}')

    surrogate_prob_field = compute_surrogate_probability_field(
        poly=surrogate_model,
        levels=levels,
        dist=distribution,
        minimum_allowable_value=minimum_allowable_value,
        convert_from_log_scale=convert_from_log_scale,
        convert_from_depths=convert_from_depths,
        depths=samples['depth'],
    )

    surrogate_prob_field = xarray.DataArray(
        surrogate_prob_field,
        coords={
            'level': levels,
            **{
                coord: values
                for coord, values in samples.coords.items()
                if coord not in ['run', 'type']
            },
        },
        dims=('level', *(dim for dim in samples.dims if dim not in ['run', 'type'])),
    )

    return surrogate_prob_field


def probability_field_from_surrogate(
    levels: List[float],
    surrogate_model: numpoly.ndpoly,
    distribution: chaospy.Distribution,
    training_set: xarray.Dataset,
    minimum_allowable_value: float = None,
    convert_from_log_scale: Union[bool, float] = False,
    convert_from_depths: Union[bool, float] = False,
    element_table: xarray.DataArray = None,
    filename: PathLike = None,
) -> xarray.Dataset:

    if filename is not None and not isinstance(filename, Path):
        filename = Path(filename)

    if filename is None or not filename.exists():
        surrogate_prob_field = probability_field_from_samples(
            samples=training_set,
            levels=levels,
            surrogate_model=surrogate_model,
            distribution=distribution,
            minimum_allowable_value=minimum_allowable_value,
            convert_from_log_scale=convert_from_log_scale,
            convert_from_depths=convert_from_depths,
        )

        # before evaluating prob. field for model set null water elevation to the ground elevation
        # training_set = numpy.fmax(training_set, -training_set['depth'])
        if minimum_allowable_value is not None:
            too_small = (training_set + training_set['depth']).values < minimum_allowable_value
            training_set.values[too_small] = numpy.nan

        ds1, ds2 = xarray.broadcast(training_set, surrogate_prob_field['level'])
        modeled_prob_field = (ds1 >= ds2).sum(dim='run') / len(training_set.run)

        node_prob_field = xarray.combine_nested(
            [surrogate_prob_field, modeled_prob_field], concat_dim='source'
        ).assign_coords(source=['surrogate', 'model'])

        node_prob_field = node_prob_field.to_dataset(name='probabilities')

        node_prob_field = node_prob_field.assign(
            differences=numpy.fabs(surrogate_prob_field - modeled_prob_field)
        )

        if element_table is not None:
            node_prob_field = node_prob_field.assign_coords({'element': element_table})

        if filename is not None:
            LOGGER.info(f'saving prob_field to "{filename}"')
            node_prob_field.to_netcdf(filename)
    else:
        LOGGER.info(f'loading prob_field from "{filename}"')
        node_prob_field = xarray.open_dataset(filename)

    return node_prob_field


def compute_surrogate_probability_field(
    poly: numpoly.ndpoly,
    levels: List[float],
    dist: chaospy.Distribution,
    sample: int = 10000,
    minimum_allowable_value: float = None,
    convert_from_log_scale: Union[bool, float] = False,
    convert_from_depths: Union[bool, float] = False,
    depths: xarray.DataArray = None,
    **kws,
):

    poly = chaospy.aspolynomial(poly)
    shape = poly.shape
    poly = poly.ravel()

    levels = numpy.asarray(levels).ravel()
    dim = len(dist)

    # Interior
    Z = dist.sample(sample, **kws).reshape(len(dist), sample)
    poly1 = poly(*Z)

    # Min/max
    ext = numpy.mgrid[(slice(0, 2, 1),) * dim].reshape(dim, 2 ** dim).T
    ext = numpy.where(ext, dist.lower, dist.upper).T
    poly2 = poly(*ext)
    poly2 = numpy.array([_ for _ in poly2.T if not numpy.any(numpy.isnan(_))]).T

    # Finish
    if poly2.shape:
        poly1 = numpy.concatenate([poly1, poly2], -1)
    if isinstance(convert_from_log_scale, float):
        poly1 = convert_from_log_scale ** poly1
    elif convert_from_log_scale:
        poly1 = numpy.exp(poly1)
    samples = poly1.shape[1]

    # adjustments and elev corrections
    if isinstance(convert_from_depths, (float, numpy.ndarray)):
        poly1 -= convert_from_depths
    if minimum_allowable_value is not None:
        too_small = poly1 < minimum_allowable_value
        poly1[too_small] = numpy.nan
    if isinstance(convert_from_depths, (float, numpy.ndarray)) or convert_from_depths:
        # TODO: Sanity check for depth vs poly shapes
        poly1 -= depths.values[:, None]

    out = (poly1[:, :, None] > (levels[None, None, :])).sum(axis=1).T / samples

    out = out.reshape(levels.shape + shape)

    return out



from abc import ABC, abstractmethod
from functools import lru_cache
from os import PathLike
from pathlib import Path
from typing import List

from geopandas import GeoDataFrame
from matplotlib import cm, colors, pyplot
from matplotlib.lines import Line2D
import numpy
import pandas
from pandas import DataFrame
from pyproj import CRS, Geod
import shapely

from ensembleperturbation.parsing.adcirc import (
    ElevationStationOutput,
    parse_adcirc_outputs,
    VelocityStationOutput,
)
from ensembleperturbation.utilities import get_logger

LOGGER = get_logger('parsing.comparison')

OBSERVATION_COLOR_MAP = cm.get_cmap('Blues')
MODEL_COLOR_MAP = cm.get_cmap('Reds')
ERROR_COLOR_MAP = cm.get_cmap('prism')

LINESTYLES = {'coldstart': ':', 'hotstart': '-'}

ADCIRC_VARIABLES = DataFrame(
    {
        'stations': ['fort.61.nc', 'fort.62.nc', 'fort.62.nc'],
        'model': ['fort.63.nc', 'fort.64.nc', 'fort.64.nc'],
        'max': ['maxele.63.nc', 'maxvel.63.nc', 'maxvel.63.nc'],
        'unit': ['m', 'm/s', 'm/s'],
        'name': ['zeta', 'u-vel', 'v-vel'],
    },
    index=['zeta', 'u', 'v'],
)


class ModelReferenceComparison(ABC):
    """
    abstraction of a comparison between reference data and modeled data
    """

    def __init__(
        self, input_directory: str, output_directory: str, variables: List[str],
    ):
        """
        :param input_directory: directory containing model inputs
        :param output_directory: directory containing model outputs
        :param variables: model variables to compare
        """

        if not isinstance(input_directory, Path):
            input_directory = Path(input_directory)
        if not isinstance(output_directory, Path):
            output_directory = Path(output_directory)

        self.input_directory = input_directory
        self.output_directory = output_directory
        self.variables = variables

    @property
    @abstractmethod
    def values(self) -> GeoDataFrame:
        raise NotImplementedError

    @abstractmethod
    def plot_errors(self, show: bool = False):
        raise NotImplementedError

    @property
    @abstractmethod
    def rmses(self) -> dict:
        raise NotImplementedError


class ADCIRCReferenceComparison(ModelReferenceComparison):
    def __init__(
        self,
        input_directory: str,
        output_directory: str,
        variables: List[str],
        stages: List[str] = None,
    ):
        """
        :param input_directory: directory containing model inputs
        :param output_directory: directory containing model outputs
        :param variables: model variables to compare
        :param stages: ADCIRC run stage (``coldstart``, ``hotstart``)
        """

        super(ADCIRCReferenceComparison, self).__init__(
            input_directory=input_directory,
            output_directory=output_directory,
            variables=variables,
        )

        self.stages = stages if stages is not None else ['coldstart', 'hotstart']

        self.fort14_filename = input_directory / 'fort.14'
        self.fort15_filename = input_directory / 'fort.15'

        self.runs = parse_adcirc_outputs(output_directory)

        self.crs = CRS.from_epsg(4326)
        ellipsoid = self.crs.datum.to_json_dict()['ellipsoid']
        self.geodetic = Geod(
            a=ellipsoid['semi_major_axis'], rf=ellipsoid['inverse_flattening']
        )

        first_run_stage = self.runs[list(self.runs)[0]][self.stages[0]]

        stations_basename = ADCIRC_VARIABLES.loc[self.variables[0]]['stations']
        if stations_basename in first_run_stage:
            self.stations = first_run_stage[stations_basename]
        else:
            self.stations = GeoDataFrame({'name': []}, geometry=[])

        self.mesh = first_run_stage[ADCIRC_VARIABLES.loc[self.variables[0]]['max']]

        self.stations = self.stations[
            self.stations.within(self.mesh.unary_union.convex_hull.buffer(0.01))
        ]
        self.stations.reset_index(drop=True, inplace=True)

    @property
    @lru_cache(maxsize=1)
    def station_mesh_vertices(self):
        nearest_mesh_vertices = []
        for station_index, station in self.stations.iterrows():
            nearest_mesh_vertex = shapely.ops.nearest_points(
                station.geometry, self.mesh.unary_union
            )[1]

            distance = self.geodetic.line_length(
                [station.geometry.x, nearest_mesh_vertex.x],
                [station.geometry.y, nearest_mesh_vertex.y],
            )
            mesh_index = self.mesh.cx[
                nearest_mesh_vertex.x, nearest_mesh_vertex.y
            ].index.item()
            nearest_mesh_vertices.append(
                GeoDataFrame(
                    {
                        'station': station['name'],
                        'station_x': station.geometry.x,
                        'station_y': station.geometry.y,
                        'distance': distance,
                    },
                    geometry=[nearest_mesh_vertex],
                    index=[mesh_index],
                )
            )
        if len(nearest_mesh_vertices) == 0:
            nearest_mesh_vertices.append(
                GeoDataFrame(
                    {'station': [], 'station_x': [], 'station_y': [], 'distance': [],},
                    geometry=[],
                    index=[],
                )
            )
        nearest_mesh_vertices = pandas.concat(nearest_mesh_vertices)
        nearest_mesh_vertices.reset_index(drop=True, inplace=True)
        return nearest_mesh_vertices

    @property
    @lru_cache(maxsize=1)
    def values(self) -> GeoDataFrame:
        observed_values = []
        for stage in self.stages:
            stations_filename = (
                self.output_directory
                / list(self.runs)[0]
                / stage
                / ADCIRC_VARIABLES.loc[self.variables[0]]['stations']
            )
            if stations_filename.exists():

                observed_values.append(
                    self.parse_stations(stations_filename, self.stations['name'])
                )
            else:
                LOGGER.warning(f'stations file not found at "{stations_filename}"')

        if len(observed_values) == 0:
            raise NoDataError('no station reference data provided')

        observed_values = pandas.concat(observed_values)

        values = []
        model_output_basename = ADCIRC_VARIABLES.loc[self.variables[0]]['model']
        for nearest_mesh_index, nearest_mesh_vertex in self.station_mesh_vertices.iterrows():
            station_name = nearest_mesh_vertex['station']

            station_values = None
            for run_name, stages in self.runs.items():
                run_values = []
                for stage, datasets in stages.items():
                    model_times = datasets[model_output_basename]['time']
                    modeled_values = {
                        variable_name: datasets[model_output_basename]['data'][
                            ADCIRC_VARIABLES.loc[variable_name]['name']
                        ]
                        for variable_name in self.variables
                    }

                    run_values.append(
                        GeoDataFrame(
                            {
                                'stage': stage,
                                'time': model_times,
                                'distance': nearest_mesh_vertex['distance'],
                                'geometry': [
                                    nearest_mesh_vertex.geometry for _ in model_times
                                ],
                                **{
                                    variable_name: variable_data[:, nearest_mesh_index]
                                    for variable_name, variable_data in modeled_values.items()
                                },
                            }
                        )
                    )

                    del model_times, modeled_values
                run_values = pandas.concat(run_values)
                run_values.columns = [
                    f'{run_name}_{column}' if column in self.variables else column
                    for column in run_values.columns
                ]

                if station_values is None:
                    station_values = run_values
                else:
                    station_run_values = run_values[
                        [
                            'stage',
                            'time',
                            'distance',
                            'geometry',
                            *(f'{run_name}_{variable}' for variable in self.variables),
                        ]
                    ]
                    station_values = pandas.merge(
                        station_values,
                        station_run_values,
                        how='left',
                        on=['stage', 'time', 'distance', 'geometry'],
                    )

            station_observed_values = observed_values[
                observed_values['station'] == station_name
            ]
            station_observed_values = station_observed_values[['time', *self.variables]]
            station_observed_values.columns = [
                'time',
                *(f'observed_{variable}' for variable in self.variables),
            ]

            station_values = pandas.merge(
                station_values, station_observed_values, how='left', on='time'
            )

            station_values.insert(0, 'station', station_name)
            values.append(station_values)

        values = pandas.concat(values)
        values.sort_values('time', inplace=True)

        values = values.iloc[
            :,
            [
                *range(5),
                *range(-len(self.variables), 0),
                *range(5, len(values.columns) - len(self.variables)),
            ],
        ]
        values.reset_index(drop=True, inplace=True)
        return values

    @property
    @lru_cache(maxsize=1)
    def errors(self) -> GeoDataFrame:
        values = self.values
        observed_values = values.iloc[:, [0, 2, *range(5, 5 + len(self.variables))]]

        errors = []
        for _, station in self.stations.iterrows():
            station_modeled_values = values[values['station'] == station['name']]
            station_distance = pandas.unique(station_modeled_values['distance'])[0]

            station_observed_values = observed_values[
                observed_values['station'] == station['name']
            ].iloc[:, 1:]
            station_observed_values.columns = ['time', *self.variables]

            station_errors = None
            for run_name in self.runs:
                run_modeled_values = station_modeled_values[
                    ['time', *(column for column in values.columns if run_name in column)]
                ]
                run_modeled_values.columns = ['time', *self.variables]

                run_errors = run_modeled_values - station_observed_values
                del run_modeled_values

                run_errors.columns = ['time_difference', *self.variables]
                run_errors.columns = [
                    f'{run_name}_{column}' if column in self.variables else column
                    for column in run_errors.columns
                ]

                run_errors.insert(0, 'time', station_modeled_values['time'])
                run_errors.insert(1, 'stage', station_modeled_values['stage'])
                run_errors.insert(2, 'distance', station_distance)

                if station_errors is None:
                    station_errors = run_errors
                else:
                    run_errors = run_errors[
                        ['time', *(f'{run_name}_{variable}' for variable in self.variables)]
                    ]
                    station_errors = pandas.merge(
                        station_errors, run_errors, how='left', on=['time']
                    )

            station_errors.insert(0, 'station', station['name'])
            errors.append(station_errors)

        errors = pandas.concat(errors)
        errors.sort_values('time', inplace=True)
        errors.reset_index(drop=True, inplace=True)
        return errors

    @property
    @lru_cache(maxsize=1)
    def rmses(self) -> dict:
        errors = self.errors
        rmses = []
        for _, station in self.stations.iterrows():
            station_errors = errors[errors['station'] == station['name']]
            station_distance = pandas.unique(station_errors['distance'])[0]
            station_rmses = []
            for run_name in self.runs:
                run_errors = station_errors[
                    [
                        'stage',
                        'time',
                        *(column for column in errors.columns if run_name in column),
                    ]
                ]
                run_rmses = []
                for stage in self.stages:
                    stage_errors = run_errors[run_errors['stage'] == stage]
                    stage_rmses = {'stage': stage}
                    for column in stage_errors.iloc[:, 2:]:
                        variable = column.replace(f'{run_name}_', '')
                        stage_rmses[variable] = numpy.sqrt((stage_errors[column] ** 2).mean())
                    run_rmses.append(stage_rmses)
                run_rmses = DataFrame.from_dict(
                    dict(zip(range(len(run_rmses)), run_rmses)), orient='index'
                )
                run_rmses.insert(0, 'run', run_name)
                station_rmses.append(run_rmses)

            station_rmses = pandas.concat(station_rmses)
            station_rmses = GeoDataFrame(
                station_rmses, geometry=[station.geometry for _ in range(len(station_rmses))]
            )

            station_rmses.insert(1, 'station', station['name'])
            station_rmses.insert(1, 'distance', station_distance)
            rmses.append(station_rmses)

        rmses = pandas.concat(rmses)
        rmses.reset_index(drop=True, inplace=True)
        return rmses

    @abstractmethod
    def parse_stations(self, filename: PathLike, station_names: List[str]) -> GeoDataFrame:
        raise NotImplementedError

    def plot_values(self, show: bool = False):
        values = self.values

        run_index_normalizer = colors.Normalize(0, len(self.runs))

        figure = pyplot.figure()
        figure.suptitle('stations')
        sharing_axis = None

        value_axes = {}
        for station_index, station in self.stations.iterrows():
            station_axes = {}
            for variable_index, variable in enumerate(self.variables):
                axis = figure.add_subplot(
                    len(self.stations) * len(self.variables),
                    1,
                    station_index * len(self.variables) + (variable_index + 1),
                    sharex=sharing_axis,
                )
                if sharing_axis is None:
                    sharing_axis = axis
                station_axes[variable] = axis

            value_axes[station['name']] = station_axes

        for station_index, station in self.stations.iterrows():
            axes = value_axes[station['name']]
            station_values = values[values['station'] == station['name']]
            for stage in self.stages:
                stage_values = station_values[station_values['stage'] == stage]
                observed_values = stage_values[
                    [
                        'time',
                        *(column for column in stage_values.columns if 'observed_' in column),
                    ]
                ]
                observed_values.columns = ['time', *self.variables]

                for run_index, run_name in enumerate(self.runs):
                    observation_color = OBSERVATION_COLOR_MAP(run_index_normalizer(run_index))
                    model_color = MODEL_COLOR_MAP(run_index_normalizer(run_index))

                    modeled_values = stage_values[
                        [
                            'time',
                            *(
                                column
                                for column in stage_values.columns
                                if f'{run_name}_' in column
                            ),
                        ]
                    ]
                    modeled_values.columns = ['time', *self.variables]

                    for variable, axis in axes.items():
                        axis.plot(
                            observed_values['time'],
                            observed_values[variable],
                            color=observation_color,
                            linestyle=LINESTYLES[stage],
                        )
                        axis.plot(
                            modeled_values['time'],
                            modeled_values[variable],
                            color=model_color,
                            linestyle=LINESTYLES[stage],
                        )

        handles = [
            Line2D([0], [0], color='b', label='observation'),
            Line2D([0], [0], color='r', label='model'),
            Line2D([0], [0], color='k', linestyle=LINESTYLES['coldstart'], label='coldstart'),
            Line2D([0], [0], color='k', linestyle=LINESTYLES['hotstart'], label='hotstart'),
        ]

        for station_name, axes in value_axes.items():
            for variable, axis in axes.items():
                axis.set_title(f'station {station_name} {variable}', loc='left')
                axis.hlines([0], *axis.get_xlim(), color='k', linestyle='--')
                axis.set_ylabel(f'{variable} ({ADCIRC_VARIABLES.loc[variable]["unit"]})')
                axis.legend(handles=handles)

        if show:
            pyplot.show()

    def plot_errors(self, show: bool = False):
        errors = self.errors

        station_index_normalizer = colors.Normalize(0, len(self.stations))

        figure = pyplot.figure()
        figure.suptitle('errors')
        sharing_axis = None

        error_axes = {}
        for variable_index, variable in enumerate(self.variables):
            axis = figure.add_subplot(
                len(self.variables), 1, variable_index + 1, sharex=sharing_axis
            )
            if sharing_axis is None:
                sharing_axis = axis

            error_axes[variable] = axis

        for variable, axis in error_axes.items():
            for station_index, station in self.stations.iterrows():
                error_color = ERROR_COLOR_MAP(station_index_normalizer(station_index))

                station_errors = errors[errors['station'] == station['name']]

                for run_index, run_name in enumerate(self.runs):
                    variable_errors = station_errors[
                        [
                            'time',
                            'stage',
                            *(
                                column
                                for column in errors
                                if column == f'{run_name}_' f'{variable}'
                            ),
                        ]
                    ]
                    variable_errors.columns = ['time', 'stage', variable]
                    for stage in self.stages:
                        stage_errors = variable_errors[variable_errors['stage'] == stage]
                        axis.plot(
                            stage_errors['time'],
                            stage_errors[variable],
                            color=error_color,
                            linestyle=LINESTYLES[stage],
                        )

        handles = [
            *(
                Line2D(
                    [0],
                    [0],
                    color=ERROR_COLOR_MAP(station_index_normalizer(station_index)),
                    label=f'station {station["name"]}',
                )
                for station_index, station in self.stations.iterrows()
            ),
            Line2D([0], [0], color='k', linestyle=LINESTYLES['coldstart'], label='coldstart'),
            Line2D([0], [0], color='k', linestyle=LINESTYLES['hotstart'], label='hotstart'),
        ]

        for variable, axis in error_axes.items():
            axis.set_title(f'{variable} error', loc='left')
            axis.hlines([0], *axis.get_xlim(), color='k', linestyle='--')
            axis.set_ylabel(f'{variable} error ' f'({ADCIRC_VARIABLES.loc[variable]["unit"]})')
            axis.legend(handles=handles)

        if show:
            pyplot.show()

    def plot_rmse(self, show: bool = False):
        rmses = self.rmses

        station_index_normalizer = colors.Normalize(0, len(self.stations))

        figure = pyplot.figure()
        figure.suptitle('RMSE')
        sharing_axis = None

        axes = {}
        for variable_index, variable in enumerate(self.variables):
            axis = figure.add_subplot(len(self.variables), 1, variable_index + 1)
            if sharing_axis is None:
                sharing_axis = axis
            axes[variable] = axis

        for variable in self.variables:
            variable_rmses = rmses[
                [
                    'run',
                    'station',
                    'stage',
                    *(column for column in rmses.columns if column == variable),
                ]
            ]

            axis = axes[variable]
            for stage in self.stages:
                stage_rmses = variable_rmses[variable_rmses['stage'] == stage]
                for station_index, station in self.stations.iterrows():
                    station_rmses = stage_rmses[stage_rmses['station'] == station['name']]
                    axis.plot(
                        station_rmses['run'],
                        station_rmses[variable],
                        color=ERROR_COLOR_MAP(station_index_normalizer(station_index)),
                        linestyle=LINESTYLES[stage],
                        label=f'{stage} {variable}',
                    )

        handles = [
            *(
                Line2D(
                    [0],
                    [0],
                    color=ERROR_COLOR_MAP(station_index_normalizer(station_index)),
                    label=f'station {station["name"]}',
                )
                for station_index, station in self.stations.iterrows()
            ),
            Line2D([0], [0], color='k', linestyle=LINESTYLES['coldstart'], label='coldstart'),
            Line2D([0], [0], color='k', linestyle=LINESTYLES['hotstart'], label='hotstart'),
        ]

        for variable, axis in axes.items():
            axis.set_xlabel('run')
            axis.set_ylabel(f'{variable} ({ADCIRC_VARIABLES.loc[variable]["unit"]})')
            axis.legend(handles=handles)

        if show:
            pyplot.show()


class ZetaComparison(ADCIRCReferenceComparison):
    """
    comparison between reference and ADCIRC-modeled sea-surface elevation
    """

    def __init__(self, input_directory: str, output_directory: str):
        """
        :param input_directory: directory containing model inputs
        :param output_directory: directory containing model outputs
        """

        super().__init__(
            input_directory, output_directory, ['zeta'], ['coldstart', 'hotstart']
        )

    def parse_stations(self, filename: PathLike, station_names: List[str]) -> GeoDataFrame:
        return ElevationStationOutput.read(filename, station_names)


class VelocityComparison(ADCIRCReferenceComparison):
    """
    comparison between reference and ADCIRC-modeled sea-surface horizontal velocity
    """

    def __init__(self, input_directory: str, output_directory: str):
        """
        :param input_directory: directory containing model inputs
        :param output_directory: directory containing model outputs
        """

        super().__init__(
            input_directory, output_directory, ['u', 'v'], ['coldstart', 'hotstart']
        )

    def parse_stations(self, filename: PathLike, station_names: List[str]) -> GeoDataFrame:
        return VelocityStationOutput.read(filename, station_names)


class NoDataError(Exception):
    pass


def insert_magnitude_components(
    dataframe: DataFrame,
    u: str = 'u',
    v: str = 'v',
    magnitude: str = 'magnitude',
    direction: str = 'direction',
    velocity_index: int = None,
    direction_index: int = None,
):
    if velocity_index is None:
        velocity_index = len(dataframe.columns)
    if direction_index is None:
        direction_index = velocity_index + 1
    dataframe.insert(velocity_index, magnitude, numpy.hypot(dataframe[u], dataframe[v]))
    dataframe.insert(direction_index, direction, numpy.arctan2(dataframe[u], dataframe[v]))



from abc import ABC, abstractmethod
from enum import Enum
from os import PathLike
from pathlib import Path
import pickle
from typing import Any, Collection, Dict, List, Mapping, Tuple, Union

import dask
import geopandas
from geopandas import GeoDataFrame
import numpy
import pandas
from pandas import DataFrame
from pyproj.transformer import Transformer
from scipy.spatial import KDTree
from shapely.geometry import Point
from stormevents.nhc import VortexTrack
from typepigeon import convert_value
import xarray
from xarray import DataArray, Dataset

from ensembleperturbation.perturbation.atcf import parse_vortex_perturbations
from ensembleperturbation.utilities import get_logger

LOGGER = get_logger('parsing.adcirc')


class ElevationSelection(Enum):
    wet = 'wet'
    inundated = 'inundated'
    dry = 'dry'


class AdcircOutput(ABC):
    filename: PathLike
    variables: List[str]
    drop_variables: List[str] = ['neta', 'nvel', 'max_nvdll', 'max_nvell']
    nodata: float = -99999.0

    @classmethod
    @abstractmethod
    def read(cls, filename: PathLike, names: List[str] = None) -> Union[DataFrame, DataArray]:
        raise NotImplementedError

    @classmethod
    def async_read(
        cls, filename: PathLike, pickle_filename: PathLike, variables: List[str] = None
    ) -> (Path, Path):
        if not isinstance(filename, Path):
            filename = Path(filename)
        if not isinstance(pickle_filename, Path):
            pickle_filename = Path(pickle_filename)

        data = cls.read(filename=filename, names=variables)
        pickle_filename = pickle_data(data, pickle_filename)

        return filename, pickle_filename

    @classmethod
    def read_directory(
        cls, directory: PathLike, variables: List[str] = None, parallel: bool = False
    ) -> Dataset:
        """
        Compile a dataset from output files in the given directory.

        :param directory: directory containing output files
        :param variables: variables to return
        :param parallel: load data concurrently with Dask
        :return: dataset of output data
        """

        if not isinstance(directory, Path):
            directory = Path(directory)

        if variables is None:
            variables = cls.variables

        filename_pattern = f'*/{cls.filename}'
        filenames = list(directory.glob(filename_pattern))
        if len(filenames) > 0:
            LOGGER.info(
                f'found {len(filenames)} files matching "{directory / filename_pattern}"'
            )
        else:
            raise FileNotFoundError(
                f'could not find any files matching "{directory / filename_pattern}"'
            )

        drop_variables = cls.drop_variables

        with xarray.open_dataset(
            filenames[0], drop_variables=drop_variables
        ) as sample_dataset:
            drop_variables.extend(
                variable_name
                for variable_name in sample_dataset.variables
                if variable_name not in variables
                and variable_name not in ['node', 'time', 'x', 'y', 'depth', 'element']
            )

        with dask.config.set(**{'array.slicing.split_large_chunks': True}):
            dataset = xarray.open_mfdataset(
                filenames,
                drop_variables=drop_variables,
                combine='nested',
                concat_dim=xarray.DataArray(
                    [filename.parent.name for filename in filenames], dims=['run'], name='run',
                ),
                parallel=parallel,
                lock=False,
            )

        if 'depth' in dataset:
            dataset = dataset.assign_coords(
                {'depth': dataset['depth'].isel(run=0).drop_vars('run')}
            )

        if 'element' in dataset:
            dataset = dataset.assign_coords(  # subtract one from element table
                {'element': dataset['element'].isel(run=0).drop_vars('run') - 1}
            )

        if 'node' not in dataset:
            dataset = dataset.assign_coords({'node': dataset['node']})

        if 'element' in dataset:
            return dataset[variables].assign_coords({'element': dataset['element']})
        else:
            return dataset[variables]

    @classmethod
    @abstractmethod
    def subset(
        cls,
        dataset: Union[Dataset, DataArray],
        bounds: (float, float, float, float) = None,
        **kwargs,
    ) -> Union[Dataset, DataArray]:
        raise NotImplementedError


class TimeSeriesOutput(ABC):
    pass


class StationTimeSeriesOutput(AdcircOutput, TimeSeriesOutput, ABC):
    @classmethod
    @abstractmethod
    def read(cls, filename: PathLike, names: List[str] = None) -> GeoDataFrame:
        raise NotImplementedError

    @classmethod
    def subset(
        cls,
        dataset: Union[Dataset, DataArray],
        bounds: (float, float, float, float) = None,
        **kwargs,
    ) -> Union[Dataset, DataArray]:
        subset = ~dataset['station'].isnull()

        if bounds is not None:
            LOGGER.debug(f'filtering within bounds {bounds}')
            if bounds[0] is not None:
                subset = numpy.logical_and(subset, dataset['x'] > bounds[0])
            if bounds[2] is not None:
                subset = numpy.logical_and(subset, dataset['x'] < bounds[2])
            if bounds[1] is not None:
                subset = numpy.logical_and(subset, dataset['y'] > bounds[1])
            if bounds[3] is not None:
                subset = numpy.logical_and(subset, dataset['y'] < bounds[3])

        return subset


class ElevationStationOutput(StationTimeSeriesOutput):
    """
    ``fort.61`` - Elevation Time Series at Specified Elevation Recording Stations

    https://adcirc.org/home/documentation/users-manual-v52/output-file-descriptions/elevation-time-series-specified-elevation-recording-stations-fort-61
    """

    filename = 'fort.61.nc'
    variables = ['station_name', 'zeta']

    @classmethod
    def read(cls, filename: PathLike, names: List[str] = None) -> GeoDataFrame:
        dataset = xarray.open_dataset(filename, drop_variables=cls.drop_variables)

        coordinate_variables = ['x', 'y']
        coordinates = numpy.stack(
            [dataset[variable] for variable in coordinate_variables], axis=1
        )
        times = dataset['time']

        all_station_names = [
            station_name.tobytes().decode().strip().strip("'")
            for station_name in dataset['station_name']
        ]

        stations = []
        for station_name in names:
            station_index = all_station_names.index(station_name)
            station_coordinates = coordinates[station_index]
            station_point = Point(station_coordinates[0], station_coordinates[1])

            stations.append(
                GeoDataFrame(
                    {
                        'time': times,
                        'zeta': dataset['zeta'][:, station_index],
                        'station': station_name,
                    },
                    geometry=[station_point for _ in times],
                )
            )

        return pandas.concat(stations)


class VelocityStationOutput(StationTimeSeriesOutput):
    """
    ``fort.62`` - Depth-averaged Velocity Time Series at Specified Velocity Recording Stations

    https://adcirc.org/home/documentation/users-manual-v52/output-file-descriptions/depth-averaged-velocity-time-series-specified-velocity-recording-stations-fort-62
    """

    filename = 'fort.62.nc'
    variables = ['station_name', 'u-vel', 'v-vel']

    @classmethod
    def read(cls, filename: PathLike, names: List[str] = None) -> GeoDataFrame:
        dataset = xarray.open_dataset(filename, drop_variables=cls.drop_variables)

        coordinate_variables = ['x', 'y']
        coordinates = numpy.stack(
            [dataset[variable] for variable in coordinate_variables], axis=1
        )
        times = dataset['time']

        all_station_names = [
            station_name.tobytes().decode().strip().strip("'")
            for station_name in dataset['station_name']
        ]

        stations = []
        for station_name in names:
            station_index = all_station_names.index(station_name)
            station_coordinates = coordinates[station_index]
            station_point = Point(station_coordinates[0], station_coordinates[1])

            stations.append(
                GeoDataFrame(
                    {
                        'time': times,
                        'u': dataset['u-vel'][:, station_index],
                        'v': dataset['v-vel'][:, station_index],
                        'station': station_name,
                    },
                    geometry=[station_point for _ in times],
                )
            )

        return pandas.concat(stations)


class FieldOutput(AdcircOutput, ABC):
    @classmethod
    def read(cls, filename: PathLike, names: List[str] = None) -> Union[DataFrame, DataArray]:
        """
        Parse ADCIRC output files

        :param filename: file path to ADCIRC NetCDF output
        :param names: list of data variables to extract
        :return: parsed data
        """

        if not isinstance(filename, Path):
            filename = Path(filename)

        LOGGER.debug(f'opening "{"/".join(filename.parts[-2:])}"')

        if names is None:
            for subclass in FieldOutput.__subclasses__():
                if filename.name == subclass.filename:
                    names = subclass.variables
                    break
            else:
                raise NotImplementedError(
                    f'ADCIRC output file "{filename.name}" not implemented'
                )

        dataset = xarray.open_dataset(filename, drop_variables=cls.drop_variables)

        coordinate_variables = ['x', 'y']
        if 'depth' in dataset.variables:
            coordinate_variables += ['depth']
        if 'element' in dataset.variables:
            coordinate_variables += ['element']
        coordinates = numpy.stack(
            [dataset[variable] for variable in coordinate_variables], axis=1
        )

        if filename.name in ['fort.63.nc', 'fort.64.nc']:
            data = dataset[names]
        elif filename.name in ['fort.61.nc', 'fort.62.nc']:
            data = GeoDataFrame(
                {
                    'name': [
                        station_name.tobytes().decode().strip().strip("'")
                        for station_name in dataset['station_name']
                    ],
                    'x': coordinates[:, 0],
                    'y': coordinates[:, 1],
                },
                geometry=geopandas.points_from_xy(coordinates[:, 0], coordinates[:, 1]),
            )
        else:
            variables_data = {}
            for variable_name in names:
                data_array = dataset[variable_name]
                if 'time_of' in variable_name:
                    pass
                if data_array.size > 0:
                    variables_data[variable_name] = numpy.squeeze(data_array)
                else:
                    LOGGER.debug(
                        f'array "{data_array.name}" has invalid data shape "{data_array.shape}"'
                    )
                    variables_data[variable_name] = numpy.squeeze(
                        numpy.full(
                            [
                                dimension if dimension > 0 else 1
                                for dimension in data_array.shape
                            ],
                            fill_value=cls.nodata,
                        )
                    )
            columns = dict(zip(coordinate_variables, coordinates.T))
            columns.update(variables_data)
            data = GeoDataFrame(
                columns, geometry=geopandas.points_from_xy(columns['x'], columns['y'])
            )
            data[data == cls.nodata] = numpy.nan

        LOGGER.debug(f'finished reading "{"/".join(filename.parts[-2:])}"')

        return data

    @classmethod
    def subset(
        cls,
        dataset: Union[Dataset, DataArray],
        bounds: (float, float, float, float) = None,
        wind_swath: [str, int] = None,
        maximum_depth: float = None,
        minimum_depth: float = None,
        **kwargs,
    ) -> Union[Dataset, DataArray]:
        subset = ~dataset['node'].isnull()

        if wind_swath is not None:
            cyclone = wind_swath[0]
            isotach = wind_swath[1]
            LOGGER.debug(f'filtering within {cyclone} wind swath {isotach}')
            if not isinstance(cyclone, VortexTrack):
                try:
                    cyclone = VortexTrack.from_file(cyclone)
                except FileNotFoundError:
                    cyclone = VortexTrack(cyclone)
            swath = cyclone.wind_swaths(wind_speed=isotach)
            if 'BEST' in swath:
                tracks = swath['BEST']
            elif 'OFCL' in swath:
                tracks = swath['OFCL']
            else:
                raise ValueError(
                    'Neither best or official track could be found for the specified storm'
                )

            series = pandas.Series(tracks.keys())
            latest_track = tracks[series[pandas.to_datetime(series).argmax()]]

            polygon = GeoDataFrame(index=[0], geometry=[latest_track])
            with dask.config.set(**{'array.slicing.split_large_chunks': True}):
                geometry = geopandas.points_from_xy(dataset['x'].values, dataset['y'].values)
                points = GeoDataFrame(
                    {'lon': dataset['x'].values, 'lat': dataset['y'].values},
                    geometry=geometry,
                )
                inpoly = geopandas.tools.sjoin(points, polygon, predicate='within', how='left')
            subset = numpy.logical_and(subset, ~numpy.isnan(inpoly.index_right.values))

        if bounds is not None:
            LOGGER.debug(f'filtering within bounds {bounds}')
            if bounds[0] is not None:
                subset = numpy.logical_and(subset, dataset['x'] > bounds[0])
            if bounds[2] is not None:
                subset = numpy.logical_and(subset, dataset['x'] < bounds[2])
            if bounds[1] is not None:
                subset = numpy.logical_and(subset, dataset['y'] > bounds[1])
            if bounds[3] is not None:
                subset = numpy.logical_and(subset, dataset['y'] < bounds[3])

        if maximum_depth is not None:
            LOGGER.debug(f'filtering by maximum depth {maximum_depth}')
            subset = numpy.logical_and(subset, dataset['depth'] <= maximum_depth)

        if minimum_depth is not None:
            LOGGER.debug(f'filtering by minimum depth {minimum_depth}')
            subset = numpy.logical_and(subset, dataset['depth'] >= minimum_depth)

        return subset


class MaximumElevationOutput(FieldOutput):
    """
    ``maxele.63`` - Maximum Elevation at All Nodes in the Model Grid

    https://adcirc.org/home/documentation/users-manual-v52/output-file-descriptions/global-maximum-minimum-files-model-run-maxele-63-maxvel-63-maxwvel-63-maxrs-63-minpr-63/
    """

    filename = 'maxele.63.nc'
    variables = ['zeta_max', 'time_of_zeta_max']


class MaximumVelocityOutput(FieldOutput):
    """
    ``maxvel.63`` - Maximum Speed at All Nodes in the Model Grid

    https://adcirc.org/home/documentation/users-manual-v52/output-file-descriptions/global-maximum-minimum-files-model-run-maxele-63-maxvel-63-maxwvel-63-maxrs-63-minpr-63/
    """

    filename = 'maxvel.63.nc'
    variables = ['vel_max', 'time_of_vel_max']


class MinimumSurfacePressureOutput(FieldOutput):
    """
    ``minpr.63`` - Minimum Sea-level Pressure at All Nodes in the Model Grid

    https://adcirc.org/home/documentation/users-manual-v52/output-file-descriptions/global-maximum-minimum-files-model-run-maxele-63-maxvel-63-maxwvel-63-maxrs-63-minpr-63/
    """

    filename = 'minpr.63.nc'
    variables = ['pressure_min', 'time_of_pressure_min']


class MaximumSurfaceWindOutput(FieldOutput):
    """
    ``maxwvel.63`` - Maximum Surface Wind Speed at All Nodes in the Model Grid

    https://adcirc.org/home/documentation/users-manual-v52/output-file-descriptions/global-maximum-minimum-files-model-run-maxele-63-maxvel-63-maxwvel-63-maxrs-63-minpr-63/
    """

    filename = 'maxwvel.63.nc'
    variables = ['wind_max', 'time_of_wind_max']


class MaximumSurfaceRadiationStressOutput(FieldOutput):
    """
    ``maxrs.63`` - Maximum Radiation Surface Stress at All Nodes in the Model Grid

    https://adcirc.org/home/documentation/users-manual-v52/output-file-descriptions/global-maximum-minimum-files-model-run-maxele-63-maxvel-63-maxwvel-63-maxrs-63-minpr-63/
    """

    filename = 'maxrs.63.nc'
    variables = ['radstress_max', 'time_of_radstress_max']


class HotStartOutput(FieldOutput):
    """
    ``fort.67`` - Hot Start Output

    https://adcirc.org/home/documentation/users-manual-v52/input-file-descriptions/hot-start-files-fort-67-fort-68/
    """

    filename = 'fort.67.nc'
    variables = ['zeta1', 'zeta2', 'zetad', 'u-vel', 'v-vel']


class HotStartOutput2(HotStartOutput):
    """
    ``fort.68`` - Hot Start Output

    https://adcirc.org/home/documentation/users-manual-v52/input-file-descriptions/hot-start-files-fort-67-fort-68/
    """

    filename = 'fort.68.nc'


class FieldTimeSeriesOutput(FieldOutput, TimeSeriesOutput, ABC):
    pass


class ElevationTimeSeriesOutput(FieldTimeSeriesOutput):
    """
    ``fort.63`` - Elevation Time Series at All Nodes in the Model Grid

    https://adcirc.org/home/documentation/users-manual-v52/output-file-descriptions/elevation-time-series-nodes-model-grid-fort-63
    """

    filename = 'fort.63.nc'
    variables = ['zeta']

    @classmethod
    def subset(
        cls,
        dataset: Union[Dataset, DataArray],
        bounds: (float, float, float, float) = None,
        maximum_depth: float = None,
        elevation_selection: ElevationSelection = None,
        **kwargs,
    ) -> Union[Dataset, DataArray]:
        subset = super().subset(dataset, bounds=bounds, maximum_depth=maximum_depth)

        if elevation_selection is not None:
            if not isinstance(elevation_selection, ElevationSelection):
                elevation_selection = convert_value(elevation_selection, ElevationSelection)

            dry_subset = dataset['zeta'].isnull()

            if elevation_selection == ElevationSelection.wet:
                elevation_subset = ~dry_subset.any('time')
            elif elevation_selection == ElevationSelection.inundated:
                # get all nodes that experienced inundation (were both wet and dry at any time)
                elevation_subset = dry_subset.any('time') & ~dry_subset.all('time')
            else:
                elevation_subset = dry_subset.all('time')

            if 'run' in dataset:
                elevation_subset = elevation_subset.any('run')

            subset = numpy.logical_and(subset, elevation_subset)

        return subset


class VelocityTimeSeriesOutput(FieldTimeSeriesOutput):
    """
    ``fort.64`` - Depth-averaged Velocity Time Series at All Nodes in the Model Grid

    https://adcirc.org/home/documentation/users-manual-v52/output-file-descriptions/depth-averaged-velocity-time-series-nodes-model-grid-fort-64
    """

    filename = 'fort.64.nc'
    variables = ['u-vel', 'v-vel']


class SurfacePressureTimeSeriesOutput(FieldTimeSeriesOutput):
    """
    ``fort.73`` - Sea-level Pressure Time Series at All Nodes in the Model Grid

    https://adcirc.org/home/documentation/output-file-descriptions/atmospheric-pressure-time-series-nodes-model-grid-fort-73
    """

    filename = 'fort.73.nc'
    variables = ['pressure']


class SurfaceWindTimeSeriesOutput(FieldTimeSeriesOutput):
    """
    ``fort.74`` - Surface Wind Velocity Time Series at All Nodes in the Model Grid

    https://adcirc.org/home/documentation/users-manual-v52/output-file-descriptions/wind-stress-velocity-time-series-nodes-model-grid-fort-74
    """

    filename = 'fort.74.nc'
    variables = ['windx', 'windy']


def adcirc_file_data_variables(cls: type = None) -> Dict[str, List[str]]:
    file_data_variables = {}
    if cls is None:
        cls = AdcircOutput
    for subclass in cls.__subclasses__():
        try:
            file_data_variables[subclass.filename] = subclass
        except AttributeError:
            file_data_variables.update(adcirc_file_data_variables(subclass))
    return file_data_variables


ADCIRC_FILE_OUTPUTS = adcirc_file_data_variables()


def pickle_data(data: Any, filename: PathLike) -> Path:
    if not isinstance(filename, Path):
        filename = Path(filename)

    if isinstance(data, Mapping):
        filename.mkdir(parents=True, exist_ok=True)
        for variable, variable_data in data.items():
            pickle_data(variable_data, filename / variable)
    else:
        if isinstance(data, numpy.ndarray):
            filename = filename.parent / (filename.name + '.npy')
            if isinstance(data, numpy.ma.MaskedArray):
                data.dump(filename)
            else:
                numpy.save(filename, data)
        elif isinstance(data, DataFrame):
            filename = filename.parent / (filename.name + '.df')
            data.to_pickle(filename)
        else:
            filename = filename.parent / (filename.name + '.pickle')
            with open(filename, 'wb') as pickle_file:
                pickle.dump(data, pickle_file)

    return filename


def parse_adcirc_outputs(
    directory: PathLike = None, file_outputs: List[str] = None, parallel: bool = False,
) -> Dict[str, dict]:
    """
    Parse output from multiple ADCIRC runs.

    :param directory: directory containing run output directories
    :param file_outputs: output files to parse
    :param parallel: load data concurrently with Dask
    :return: variables to parsed data
    """

    if directory is None:
        directory = Path.cwd()
    elif not isinstance(directory, Path):
        directory = Path(directory)

    if file_outputs is None:
        file_outputs = ADCIRC_FILE_OUTPUTS
    elif isinstance(file_outputs, Collection):
        file_outputs = {filename: ADCIRC_FILE_OUTPUTS[filename] for filename in file_outputs}
    elif isinstance(file_outputs, Mapping):
        file_outputs = {
            filename: subclass if subclass is not None else ADCIRC_FILE_OUTPUTS[filename]
            for filename, subclass in file_outputs.items()
        }

    for basename, output_class in file_outputs.items():
        if issubclass(output_class, AdcircOutput):
            file_outputs[basename] = output_class

    output_tree = {}
    for basename, output_class in file_outputs.items():
        try:
            output_tree[basename] = output_class.read_directory(
                directory, variables=output_class.variables, parallel=parallel,
            )
        except (ValueError, FileNotFoundError) as error:
            LOGGER.warning(error)

    return output_tree


def combine_outputs(
    directory: PathLike = None,
    file_data_variables: Dict[str, List[str]] = None,
    bounds: Tuple[float, float, float, float] = None,
    maximum_depth: float = None,
    elevation_selection: ElevationSelection = None,
    output_directory: PathLike = None,
    parallel: bool = False,
) -> Dict[str, DataFrame]:
    if directory is None:
        directory = Path.cwd()
    elif not isinstance(directory, Path):
        directory = Path(directory)

    if maximum_depth is not None and not isinstance(maximum_depth, float):
        maximum_depth = float(maximum_depth)

    if output_directory is not None:
        if not isinstance(output_directory, Path):
            output_directory = Path(output_directory)
        if not output_directory.exists():
            output_directory.mkdir(parents=True, exist_ok=True)

    runs_directory = directory / 'runs'
    if not runs_directory.exists():
        raise FileNotFoundError(f'runs directory does not exist at "{runs_directory}"')

    track_directory = directory / 'track_files'
    if not track_directory.exists():
        raise FileNotFoundError(f'track directory does not exist at "{track_directory}"')

    if file_data_variables is None:
        file_data_variables = ADCIRC_FILE_OUTPUTS
    elif isinstance(file_data_variables, Collection):
        file_data_variables = {
            filename: ADCIRC_FILE_OUTPUTS[filename] for filename in file_data_variables
        }
    elif isinstance(file_data_variables, Mapping):
        file_data_variables = {
            filename: variables if variables is not None else ADCIRC_FILE_OUTPUTS[filename]
            for filename, variables in file_data_variables.items()
        }

    # parse all the inputs using built-in parser
    output_data = {'perturbations.nc': parse_vortex_perturbations(track_directory)}

    # parse all the outputs using built-in parser
    LOGGER.info(f'parsing from "{directory}"')
    parsed_files = parse_adcirc_outputs(
        directory=runs_directory, file_outputs=file_data_variables, parallel=parallel,
    )

    elevation_time_series_filename = 'fort.63.nc'
    if elevation_selection is not None:
        if elevation_time_series_filename in parsed_files:
            output_data[elevation_time_series_filename] = parsed_files[
                elevation_time_series_filename
            ]
            del parsed_files[elevation_time_series_filename]
        else:
            raise ValueError(
                f'elevation time series "{elevation_time_series_filename}" not found'
            )

    output_data.update(parsed_files)

    # generate subset
    elevation_subset = None
    for basename, file_data in output_data.items():
        if 'node' in file_data:
            num_nodes = len(file_data['node'])

            variable_shape_string = ', '.join(
                f'"{name}" {variable.shape}' for name, variable in file_data.items()
            )
            LOGGER.info(
                f'found {len(file_data)} variable(s) in "{basename}": {variable_shape_string}'
            )

            file_data_variable = file_data_variables[basename]

            subset = ~file_data['node'].isnull()

            if elevation_subset is not None:
                subset = numpy.logical_and(subset, elevation_subset)

            subset = numpy.logical_and(
                subset,
                file_data_variable.subset(
                    file_data,
                    bounds=bounds,
                    maximum_depth=maximum_depth,
                    elevation_selection=elevation_selection,
                ),
            )

            if subset is not None:
                file_data = file_data.sel(node=subset)

                LOGGER.info(
                    f'subsetted {len(file_data["node"])} out of {num_nodes} total nodes ({len(file_data["node"]) / num_nodes:3.2%})'
                )

                if elevation_selection is not None:
                    elevation_subset = subset

                if 'element' in file_data:
                    # keep only elements where all nodes are present
                    elements = file_data['element'].values
                    element_mask = numpy.isin(elements, file_data['node'].values).all(axis=1)
                    elements = elements[element_mask]
                    # map nodes in element table to local numbering system (start at 0)
                    node_mapper = numpy.zeros(file_data['node'].max().values + 1, dtype=int)
                    node_index = numpy.arange(len(file_data['node']))
                    node_mapper[file_data['node'].values] = node_index
                    elements = node_mapper[elements]
                    # update element table in dataset
                    ele_da = DataArray(data=elements, dims=['nele', 'nvertex'])
                    file_data = file_data.drop('element').assign_coords({'element': ele_da})

            output_data[basename] = file_data

    for basename, file_data in output_data.items():
        if output_directory is not None:
            output_filename = output_directory / basename
            LOGGER.info(f'writing to "{output_filename}"')
            file_data.to_netcdf(
                output_filename,
                # encoding={
                #     variable_name: {'zlib': True} for variable_name in file_data.variables
                # },
            )

    return output_data


def subset_dataset(
    ds: Dataset,
    variable: str,
    maximum_depth: float = None,
    minimum_depth: float = None,
    wind_swath: list = None,
    bounds: (float, float, float, float) = None,
    node_status_selection: dict = None,
    point_spacing: int = None,
    output_filename: PathLike = None,
):
    with dask.config.set(**{'array.slicing.split_large_chunks': True}):
        if node_status_selection is None:
            node_status_mask = None
        elif node_status_selection['mask'] == 'sometimes_wet':
            node_status_mask = (
                ~ds[variable].sel(run=node_status_selection['runs']).isnull().all('run'),
            )
        elif node_status_selection['mask'] == 'always_wet':
            node_status_mask = (
                ~ds[variable].sel(run=node_status_selection['runs']).isnull().any('run'),
            )
        else:
            raise ValueError(
                f'node_status_selection {node_status_selection["mask"]} unrecognized'
            )

        node_subset_mask = (
            FieldOutput.subset(
                ds['node'],
                maximum_depth=maximum_depth,
                minimum_depth=minimum_depth,
                bounds=bounds,
                wind_swath=wind_swath,
            ),
        )
        if node_status_mask is None:
            node_status_mask = node_subset_mask
        subsetted_nodes = ds['node'].values[
            numpy.logical_and(node_status_mask, node_subset_mask).squeeze()
        ]
        if point_spacing is not None:
            subsetted_nodes = subsetted_nodes[::point_spacing]
        subset = ds.sel(node=subsetted_nodes)

        # adjust element table if present
        if 'element' in subset:
            # keep only elements where all nodes are present
            elements = subset['element'].values
            element_mask = numpy.isin(elements, subset['node'].values).all(axis=1)
            elements = elements[element_mask]
            # map nodes in element table to local numbering system (start at 0)
            node_mapper = numpy.zeros(subset['node'].max().values + 1, dtype=int)
            node_index = numpy.arange(len(subset['node']))
            node_mapper[subset['node'].values] = node_index
            elements = node_mapper[elements]
            # update element table in dataset
            ele_da = DataArray(data=elements, dims=['nele', 'nvertex'])
            subset = subset.drop('element').assign_coords({'element': ele_da})
        try:
            subset = subset.drop_sel(run='original')
        except:
            pass
        if len(subset['node']) != len(ds['node']):
            LOGGER.info(
                f'subsetted down to {len(subset["node"])} nodes ({len(subset["node"]) / len(ds["node"]):.1%})'
            )
        if output_filename is not None:
            if not isinstance(output_filename, Path):
                output_filename = Path(output_filename)
            LOGGER.info(f'saving subset to "{output_filename}"')
            subset.to_netcdf(output_filename)

    return subset


def extrapolate_water_elevation_to_dry_areas(
    da: DataArray,
    k_neighbors: int = 1,
    idw_order: int = 1,
    compute_headloss: bool = False,
    mann_coef: float = 0.05,
    u_ref: float = 0.4,
    d_ref: float = 1.0,
    min_depth: float = 0.0,
):
    # return a deep copy of original datarrary on output
    da_adjusted = da.copy(deep=True)

    # Get coordinates in conformal projection (e.g,, Mercator)
    # for determining closest distance
    crs_from = 'EPSG:4326'  # WGS84
    crs_to = 'EPSG:3857'  # Mercator
    transformer = Transformer.from_crs(crs_from=crs_from, crs_to=crs_to, always_xy=True)
    x, y = transformer.transform(da['x'].values, da['y'].values)
    projected_coordinates = numpy.vstack([x, y]).T

    # for mapping back to node numbers
    nodes = numpy.arange(da.sizes['node'])

    # compute the friction factor for headloss calculation:
    # Rucker, et al. (2021). Natural Hazards.
    # https://doi.org/10.1007/s11069-021-04634-8
    if compute_headloss:
        friction_factor = (u_ref * mann_coef) ** 2 / d_ref ** (4 / 3)
    else:
        friction_factor = 0.0

    # inverse distance weighting of order `idw_order` with `k_nearest` neighbors
    for run in range(da.sizes['run']):
        null = numpy.isnan(da[run, :])
        tree = KDTree(projected_coordinates[~null])
        dd, nn = tree.query(projected_coordinates[null], k=k_neighbors)
        max_allowable_values = da['depth'][null].values + min_depth - numpy.finfo(float).eps
        if k_neighbors == 1:
            headloss = dd * friction_factor  # hydraulic friction loss
            da_adjusted[run, null] = numpy.fmin(
                da[run, nodes[~null][nn]].values - headloss, max_allowable_values
            )
        else:
            for kk in range(k_neighbors):
                weights = dd[:, kk] ** (-idw_order)
                headloss = dd[:, kk] * friction_factor  # hydraulic friction loss
                total_head = da[run, nodes[~null][nn[:, kk]]].values - headloss
                if kk == 0:
                    idw_sum = total_head * weights
                    weight_sum = weights
                else:
                    idw_sum += total_head * weights
                    weight_sum += weights
            da_adjusted[run, null] = numpy.fmin(idw_sum / weight_sum, max_allowable_values)

    return da_adjusted



from os import PathLike
from pathlib import Path
import pickle
from typing import Union

import cartopy
import geodatasets
import geopandas
import cmocean
from matplotlib import pyplot
import numpy as np
from sklearn.decomposition import PCA

from ensembleperturbation.plotting.geometry import plot_points, plot_surface
from ensembleperturbation.utilities import get_logger

LOGGER = get_logger('karhunen_loeve')


def karhunen_loeve_expansion(
    ymodel: np.ndarray,
    neig,
    method: Union[float, str] = None,
    output_directory: PathLike = None,
):
    if output_directory is not None and not isinstance(output_directory, Path):
        output_directory = Path(output_directory)

    # get the shape of the data
    nens, ngrid = ymodel.shape

    if neig is None:
        neig = ngrid
    elif isinstance(neig, int):
        neig = min(neig, ngrid)
    elif isinstance(neig, float):
        assert neig <= 1.0 and neig >= 0.0, 'specify 0.0 <= neig <= 1.0'

    if method == 'PCA':
        LOGGER.info(f'Using sklearn PCA decomposition method')
        # Using the scikit PCA (same as KL in discrete space)
        pca_obj = PCA(n_components=neig, random_state=666, whiten=True)
        pca_obj.fit(ymodel)
        kl_dict = {
            'mean_vector': pca_obj.mean_,
            'modes': pca_obj.components_,
            'eigenvalues': pca_obj.explained_variance_,
            'neig': len(pca_obj.explained_variance_),
            'samples': pca_obj.transform(ymodel),
        }

    else:
        LOGGER.info(f'Using native KL decomposition method')
        ymodel = ymodel.T

        # evaluate weights and eigen values
        mean_vector = np.mean(ymodel, axis=1)
        covariance = np.cov(ymodel)

        weights = trapezoidal_rule_weights(length=ngrid)

        eigen_values, eigen_vectors = karhunen_loeve_eigen_values(
            covariance=covariance, weights=weights
        )

        # Karhunen–Loève modes ('principal directions')
        modes = karhunen_loeve_modes(eigen_vectors=eigen_vectors, weights=weights)
        eigen_values[eigen_values < 1.0e-14] = 1.0e-14
        #
        relative_diagonal = karhunen_loeve_relative_diagonal(
            karhunen_loeve_modes=modes, eigen_values=eigen_values, covariance=covariance
        )
        #
        xi = karhunen_loeve_coefficient_samples(
            data=ymodel, eigen_values=eigen_values, eigen_vectors=eigen_vectors,
        )
        #
        # re-ordering the matrices (mode-1 first)
        xi = xi[:, ::-1]
        modes = modes[:, ::-1]
        eigen_values = eigen_values[::-1]

        # get desired modes
        if isinstance(neig, float):
            # determine neig that make up neig decimal fraction of the variance explained
            target = neig * sum(eigen_values)
            eig_sum = 0
            for neig in range(ngrid):
                eig_sum = eig_sum + eigen_values[neig]
                if eig_sum >= target:
                    break
            neig = neig + 1

        # form into KL dictionary
        kl_dict = {
            'mean_vector': mean_vector,
            'modes': modes[:, :neig].T,
            'eigenvalues': eigen_values[:neig],
            'neig': neig,
            'samples': xi[:, :neig],
        }

        # plot the eigenvalues and KL modes, and save to file
    if output_directory is not None:
        filename = output_directory / 'karhunen_loeve.pkl'
        with open(filename, 'wb') as kl_handle:
            LOGGER.info(f'saving Karhunen-Loeve expansion to "{filename}"')
            pickle.dump(kl_dict, kl_handle)

        figure = pyplot.figure()
        axis = figure.add_subplot(1, 1, 1)

        axis.plot(range(1, kl_dict['neig'] + 1), kl_dict['eigenvalues'], 'o-')

        axis.set_xlabel('x')
        axis.set_ylabel('Eigenvalue')

        figure.savefig(output_directory / 'KL_eigenvalues.png', dpi=200, bbox_inches='tight')
        pyplot.close()

    return kl_dict


def karhunen_loeve_pc_coefficients(
    kl_dict: dict, pc_coefficients: np.ndarray,
):
    """
    Get the joint karhunen_loeve Polynomial chaos polynomial coefficients
    from Eq (6) in Sargsyan, K. and Ricciuto D. (2021):
    sum_{k=0}^{K} (kroneckerdelta_{k0}*mean(f(z)) +  
                   sum_{j=0}^{L} [c_{jk}*\sqrt(ev_j)*ef_j(z)]) 
    K -> num PC coefficients
    L -> num KL modes
    c -> PC coefficient 
    ev -> eigenvalue
    ef -> eigenfunction
    """

    # get the coefficients of the PC for each point in z (spatiotemporal dimension)
    num_points = len(kl_dict['mean_vector'])
    num_modes = len(kl_dict['eigenvalues'])
    assert (
        num_modes == pc_coefficients.shape[0]
    ), 'number of kl_dict eigenvalues needs to be equal to the length of the first dimension of pc_coefficients'
    num_coefficients = pc_coefficients.shape[1]
    klpc_coefficients = np.zeros((num_points, num_coefficients))
    klpc_coefficients[:, 0] = kl_dict['mean_vector']
    for z_index in range(num_points):
        for coef_index in range(num_coefficients):
            for mode_index in range(num_modes):
                klpc_coefficients[z_index, coef_index] += (
                    pc_coefficients[mode_index, coef_index]
                    * np.sqrt(kl_dict['eigenvalues'][mode_index])
                    * kl_dict['modes'][mode_index, z_index]
                )

    return klpc_coefficients  #: size (num_points, num_coefficients)


def karhunen_loeve_prediction(
    kl_dict: dict,
    samples=None,
    actual_values=None,
    ensembles_to_plot=None,
    element_table=None,
    reference_line: bool = True,
    plot_directory: PathLike = None,
):
    """
    Evaluating the KL model prediction (against actual values if provided)
    
    """

    if plot_directory is not None and not isinstance(plot_directory, Path):
        plot_directory = Path(plot_directory)

    if samples is None:
        samples = kl_dict['samples']

    kl_prediction = kl_dict['mean_vector'] + np.dot(
        np.dot(samples, np.diag(np.sqrt(kl_dict['eigenvalues']))), kl_dict['modes']
    )

    if actual_values is not None:
        figure = pyplot.figure()
        axis = figure.add_subplot(1, 1, 1)

        # Plot to make sure kl_prediction and actual_values are close
        axis.plot(actual_values.values.T, kl_prediction.T, '.', markersize=1)

        axis.set_xlabel('actual')
        axis.set_ylabel('reconstructed')
        axis.title.set_text(f'KL fit for each ensemble')

        xlim = axis.get_xlim()
        ylim = axis.get_ylim()
        bb_min = min(xlim[0], ylim[0])
        bb_max = max(xlim[1], ylim[1])
        axis.set_xlim(bb_min, bb_max)
        axis.set_ylim(bb_min, bb_max)

        if reference_line:
            axis.plot([bb_min, bb_max], [bb_min, bb_max], '--k', alpha=0.3, zorder=-50)

        figure.savefig(
            plot_directory / f'KL_fit.png', dpi=200, bbox_inches='tight',
        )
        pyplot.close()

        bounds = np.array(
            [
                actual_values['x'].min(),
                actual_values['y'].min(),
                actual_values['x'].max(),
                actual_values['y'].max(),
            ]
        )
        vmax = np.round(actual_values.quantile(0.98), decimals=1)
        vmin = min(0.0, np.round(actual_values.quantile(0.02), decimals=1))
        sources = {'actual': actual_values, 'reconstructed': kl_prediction}
        map_crs = cartopy.crs.PlateCarree()
        for example in ensembles_to_plot:
            figure = pyplot.figure()
            figure.set_size_inches(10, 10 / 1.61803398875)
            figure.suptitle(f'KL reconstruction comparison, ensemble #{example}')
            index = 0
            for source, value in sources.items():
                index += 1
                map_axis = figure.add_subplot(2, len(sources), index, projection=map_crs)
                map_axis.title.set_text(f'{source}')
                countries = geopandas.read_file(geodatasets.get_path('naturalearth land'))
                countries.plot(color='lightgrey', ax=map_axis)
                coast = cartopy.feature.NaturalEarthFeature(
                    category='physical', scale='50m', facecolor='none', name='coastline'
                )
                map_axis.add_feature(coast, edgecolor='grey', linewidth=0.5)
                map_axis.set_xlim((bounds[0], bounds[2]))
                map_axis.set_ylim((bounds[1], bounds[3]))

                if element_table is None:
                    im = plot_points(
                        np.vstack(
                            (actual_values['x'], actual_values['y'], value[example, :])
                        ).T,
                        axis=map_axis,
                        add_colorbar=False,
                        vmax=vmax,
                        vmin=vmin,
                        s=1,
                        extend='both',
                    )
                else:
                    im = plot_surface(
                        points=np.vstack(
                            (actual_values['x'], actual_values['y'], value[example, :])
                        ).T,
                        element_table=element_table.values,
                        axis=map_axis,
                        add_colorbar=False,
                        levels=np.linspace(vmin, vmax, 25 + 1),
                        extend='both',
                        cmap=cmocean.tools.crop(cmocean.cm.diff_r, vmin, vmax, pivot=0),
                    )

            pyplot.subplots_adjust(wspace=0.02, right=0.96)
            cax = pyplot.axes([0.95, 0.55, 0.015, 0.3])
            cbar = figure.colorbar(im, extend='both', cax=cax)
            cbar.ax.set_title('[m]')

            figure.savefig(
                plot_directory / f'KL_ensemble{example}.png', dpi=200, bbox_inches='tight',
            )
            pyplot.close()

    return kl_prediction


def trapezoidal_rule_weights(length: int):
    # Set trapezoidal rule weights
    weights = np.ones(length)
    weights[[0, -1]] = 0.5
    return np.sqrt(weights)


def karhunen_loeve_eigen_values(
    covariance: np.ndarray, weights: np.ndarray
) -> (np.ndarray, np.ndarray):
    return np.linalg.eigh(np.outer(weights, weights) * covariance)


def karhunen_loeve_modes(eigen_vectors: np.ndarray, weights: np.ndarray):
    return eigen_vectors / weights.reshape(-1, 1)  # ngrid, neig


def karhunen_loeve_relative_diagonal(
    karhunen_loeve_modes: np.ndarray, eigen_values: np.ndarray, covariance: np.ndarray
):
    cumulative_sum = (
        np.cumsum((karhunen_loeve_modes[:, :-1] * np.sqrt(eigen_values[:-1])) ** 2, axis=1)
        + 0.0
    )
    diagonal = np.diag(covariance).reshape(-1, 1) + 0.0
    return cumulative_sum / diagonal


def karhunen_loeve_coefficient_samples(
    data: np.ndarray, eigen_values: np.ndarray, eigen_vectors: np.ndarray,
):
    """
    get samples for the Karhunen–Loève coefficients from the given data

    :param data: array of data of shape `ngrid x nens`
    :param eigen_values: 1D array of eigen values
    :param eigen_vectors: 2D array of eigen vectors
    :return: samples for the Karhunen–Loève coefficients
    """
    # nens, neig
    return np.dot(
        data.T - np.mean(data, axis=1),
        eigen_vectors * trapezoidal_rule_weights(data.shape[0]).reshape(-1, 1),
    ) / np.sqrt(eigen_values)



from os import PathLike
from typing import Dict

import numpy
import pandas
from pandas import DataFrame
from scipy.spatial import cKDTree
from scipy.special import ndtri


def ensemble_array(
    input_dataframe: DataFrame, output_dataframe: DataFrame
) -> (numpy.ndarray, numpy.ndarray):
    nens = len(input_dataframe)
    ngrid = len(output_dataframe)
    dim = 4

    print(f'Parameter dimensionality: {dim}')
    print(f'Ensemble size: {nens}')
    print(f'Spatial grid size: {ngrid}')

    # Convert to proper numpy (there must be a cute pandas command to do this in a line or two...)
    pinput = numpy.empty((0, dim))
    output = numpy.empty((0, ngrid))
    for iens in range(nens):
        sample_key = 'vortex_4_variable_perturbation_' + str(iens + 1)
        pinput = numpy.append(
            pinput, input_dataframe.loc[sample_key + '.json'].to_numpy().reshape(1, -1), axis=0
        )
        output = numpy.append(
            output, output_dataframe[sample_key].to_numpy().reshape(1, -1), axis=0
        )

    # Transform the uniform dimension into gaussian
    for cdx, col in enumerate(input_dataframe.columns):
        if col == 'radius_of_maximum_winds':
            pinput[:, cdx] = ndtri((pinput[:, cdx] + 1.0) / 2.0)

    print(f'Shape of parameter input: {pinput.shape}')
    print(f'Shape of model output: {output.shape}')

    return pinput, output


def sample_points_with_equal_spacing(output_dataframe: DataFrame, spacing: float):
    # get the mask indices for an equal spatial sampling of points

    # make the point tree
    points = output_dataframe[['x', 'y']].to_numpy()
    point_tree = cKDTree(points)

    # enquire
    bad_indices = numpy.empty(0, dtype=int)
    good_indices = numpy.zeros(1, dtype=int)
    bad = 0
    counter = 0
    while bad < len(points) - 1:
        counter = counter + 1
        idx_temp = point_tree.query_ball_point(points[good_indices[-1], :], spacing)
        idx_temp = numpy.setdiff1d(idx_temp, good_indices, assume_unique=True)
        bad_indices = numpy.unique(numpy.append(bad_indices, idx_temp))
        for bdx, bad in enumerate(bad_indices):
            if bad <= good_indices[-1]:
                continue
            if bdx == len(bad_indices) - 1 or bad_indices[bdx + 1] - bad > 1:
                good_indices = numpy.append(good_indices, bad + 1)
                break
    good_indices = good_indices[:-1]

    return good_indices



import os

from matplotlib import pyplot
import numpy as np


def build_pc_expansion(
    x_filename: os.PathLike = 'xdata.dat',
    y_filename: os.PathLike = 'qoi.dat',
    output_filename: str = None,
    pc_type: str = 'HG',
    poly_order: int = 5,
    lambda_regularization: int = 0,
):
    """
    builds the polynomial-chaos expansion

    regression function inputs: 
    -x "xdata filename" 
    -y "ydata filename" 
    -s "PCtype" 
    -o "Polynomial order"  
    -l "regularization lambda"
    """
    uqtk_cmd = f'regression -x {x_filename} -y {y_filename} -s {pc_type} -o {poly_order} -l {lambda_regularization} >& regr.log'
    os.system(uqtk_cmd)
    pc_coefficients = np.loadtxt('coeff.dat')
    if output_filename is not None:
        os.rename('coeff.dat', output_filename)

    return pc_coefficients


def evaluate_pc_expansion(
    x_filename: os.PathLike = 'xdata.dat',
    parameter_filename: os.PathLike = 'coeff.dat',
    output_filename: os.PathLike = None,
    pc_type: str = 'HG',
    poly_order: int = 5,
):
    """
    evaluates the polynomial-chaos expansion
    
    pce_eval function inputs: 
    -x "xdata filename" 
    -f "parameter filename" 
    -s "PC type" 
    -o "Polynomial order"  
    """
    assert x_filename == 'xdata.dat', 'x_filename needs to be xdata.dat'
    uqtk_cmd = f'pce_eval -f {parameter_filename} -s {pc_type} -o {poly_order} >& pc_eval.log'
    os.system(uqtk_cmd)
    qoi_pc = np.loadtxt('ydata.dat')
    if output_filename is not None:
        os.rename('ydata.dat', output_filename)
    return qoi_pc


def evaluate_pc_multiindex(
    multiindex_filename: os.PathLike = 'mindex.dat',
    multiindex_type: str = 'TO',
    poly_order: int = 3,
    pc_dimension: int = 1,
):
    """
    evaluates the multi-index of the PC expansion
    
    gen_mi function inputs (generates the multi-index files): 
    -x "Multiindex type" 
    -p "PC polynomial order" 
    -q "PC dimension (number of parameters)" 
    """

    # evaluate the multi-index file
    uqtk_cmd = f'gen_mi -x {multiindex_type} -p {poly_order} -q {pc_dimension} >& mi.log'
    os.system(uqtk_cmd)
    multiindex_filename = 'mindex.dat'


def evaluate_pc_sensitivity(
    parameter_filename: os.PathLike = 'coeff.dat',
    multiindex_filename: os.PathLike = 'mindex.dat',
    pc_type: str = 'HG',
):
    """
    evaluates the Sobol sensitivities of the PC expansion
    
    pce_sens function inputs (generates the Sobol sensitivity indices): 
    -x "PC type" 
    -f "PC coefficient filename"
    -m "Multiindex file"
    """

    # evaluating the sensitivities
    uqtk_cmd = (
        f'pce_sens -f {parameter_filename} -x {pc_type} -m {multiindex_filename} >& sens.log'
    )
    os.system(uqtk_cmd)
    sensitivities = {
        'main': np.loadtxt('mainsens.dat'),
        'joint': np.loadtxt('jointsens.dat'),
        'total': np.loadtxt('totsens.dat'),
    }
    return sensitivities


def evaluate_pc_distribution_function(
    parameter_filename: os.PathLike = 'coeff.dat',
    multiindex_filename: os.PathLike = 'mindex.dat',
    pc_type: str = 'HG',
    multiindex_type: str = 'TO',
    poly_order: int = 3,
    pc_dimension: int = 1,
    num_samples: int = 1000,
    pdf_bins: int = 100,
    custom_xlabel: str = None,
    figname: str = None,
):
    """
    evaluates the PDF & CDF of the surrogate output
    
    pce_rv function inputs (generates the random variable): 
    -x "PC type" 
    -n "num samples" 
    -o "PC polynomial order" 
    -p "PC dimension (number of parameters)" 
    -f "PC coefficient filename"
    -m "multi-index filename"
    -w "type of random variable"
    
    pdf_cl function inputs (generates the pdf): 
    -i "input random variable filename" (rvar.dat from pce_rv) 
    -g "number of bins in the pdf" 
    """

    # evaluating the PC-related random variables
    uqtk_cmd = f'pce_rv -x {pc_type} -n {num_samples} -o {poly_order} -p {pc_dimension} -f {parameter_filename} -m {multiindex_filename} -w PCmi >& pc_rv.log'
    os.system(uqtk_cmd)

    # evaluating the PDF of the PC expansion
    uqtk_cmd = f'pdf_cl -i rvar.dat -g {pdf_bins} >& pdf.log'
    os.system(uqtk_cmd)
    xtarget = np.loadtxt('dens.dat')[:, :-1].squeeze()
    pdf = np.loadtxt('dens.dat')[:, -1:].squeeze()
    cdf = np.cumsum(pdf) * np.diff(xtarget)[0]

    if figname is not None:
        pyplot.figure(figsize=(12, 8))

        pyplot.plot(xtarget, pdf)

        if custom_xlabel is not None:
            pyplot.xlabel(custom_xlabel)

        pyplot.ylabel('PDF')

        pyplot.suptitle(figname)
        output_filename = f'{figname}.png'
        pyplot.savefig(output_filename, bbox_inches='tight')

    # entering the PDF/CDF into a dictionary
    distribution_dict = {
        'x': xtarget,
        'pdf': pdf,
        'cdf': cdf,
    }
    return distribution_dict


def evaluate_pc_exceedance_heights(exceedance_probabilities: np.ndarray, pc_dict: dict):
    """
    Get the heights at the desired exceedance probabilities
    
    """

    return np.interp(1.0 - exceedance_probabilities, pc_dict['cdf'], pc_dict['x'])


def evaluate_pc_exceedance_probabilities(exceedance_heights: np.ndarray, pc_dict: dict):
    """
    Get the probabilities of exceedance above desired heights
    
    """

    return 1.0 - np.interp(exceedance_heights, pc_dict['x'], pc_dict['cdf'])
